# -*- coding: utf-8 -*-
''' Classes to build full console screen '''
import fcntl
import logging
import os
import struct
import termios
import time
import bisect
from collections import defaultdict
import pandas as pd

from ...common import util


def get_terminal_size():
    '''
    Gets width and height of terminal viewport
    '''
    default_size = (30, 120)
    env = os.environ

    def ioctl_gwinsz(file_d):
        '''
        Helper to get console size
        '''
        try:
            sizes = struct.unpack('hh', fcntl.ioctl(file_d, termios.TIOCGWINSZ, '1234'))
        except Exception:
            sizes = default_size
        return sizes

    sizes = ioctl_gwinsz(0) or ioctl_gwinsz(1) or ioctl_gwinsz(2)
    if not sizes:
        try:
            file_d = os.open(os.ctermid(), os.O_RDONLY)
            sizes = ioctl_gwinsz(file_d)
            os.close(file_d.fileno())
        except Exception:
            pass
    if not sizes:
        try:
            sizes = (env['LINES'], env['COLUMNS'])
        except Exception:
            sizes = default_size
    return int(sizes[1]), int(sizes[0])


def safe_div(summ, count):
    if count == 0:
        return 0
    else:
        return float(summ) / count


def str_len(n):
    return len(str(n))


def avg_from_dict(src):
    result = {}
    count = src['count']
    for k, v in src.items():
        if k != 'count':
            result[k] = safe_div(v, count)
    return result


def krutilka():
    pos = 0
    chars = "|/-\\"
    while True:
        yield chars[pos]
        pos += 1
        if pos >= len(chars):
            pos = 0


def try_color(old, new, markup):
    order = [
        markup.WHITE, markup.GREEN, markup.CYAN,
        markup.YELLOW, markup.MAGENTA, markup.RED]
    if not old:
        return new
    else:
        if order.index(old) > order.index(new):
            return old
        else:
            return new


def combine_codes(tag_data, markup):
    net_err, http_err = 0, 0
    color = ''
    net_codes = tag_data['net_code']['count']
    for code, count in sorted(net_codes.items()):
        if count > 0:
            if int(code) == 0:
                continue
            elif int(code) == 314:
                color = try_color(color, markup.MAGENTA, markup)
                net_err += count
            else:
                color = try_color(color, markup.RED, markup)
                net_err += count
    http_codes = tag_data['proto_code']['count']
    for code, count in sorted(http_codes.items()):
        if count > 0:
            if 100 <= int(code) <= 299:
                color = try_color(color, markup.GREEN, markup)
            elif 300 <= int(code) <= 399:
                color = try_color(color, markup.CYAN, markup)
            elif 400 <= int(code) <= 499:
                color = try_color(color, markup.YELLOW, markup)
                http_err += count
            elif 500 <= int(code) <= 599:
                color = try_color(color, markup.RED, markup)
                http_err += count
            else:
                color = try_color(color, markup.MAGENTA, markup)
                http_err += count
    return (net_err, http_err, color)


class TableFormatter(object):
    def __init__(self, template, delimiters, reshape_delay=5):
        self.log = logging.getLogger(__name__)
        self.template = template
        self.delimiters = delimiters
        self.default_final = '{{{:}:>{len}}}'
        self.last_reshaped = time.time()
        self.old_shape = {}
        self.reshape_delay = reshape_delay

    def __delimiter_gen(self):
        for d in self.delimiters:
            yield d
        while True:
            yield self.delimiters[-1]

    def __prepare(self, data):
        prepared = []
        shape = {}
        for line in data:
            new = {}
            for f in line:
                if f in self.template:
                    new[f] = self.template[f]['tpl'].format(line[f])
                    if f not in shape:
                        shape[f] = len(new[f])
                    else:
                        shape[f] = max(shape[f], len(new[f]))
            prepared.append(new)
        return (prepared, shape)

    def __update_shape(self, shape):
        def change_shape():
            self.last_reshaped = time.time()
            self.old_shape = shape
        if set(shape.keys()) != set(self.old_shape.keys()):
            change_shape()
            return shape
        else:
            for f in shape:
                if shape[f] > self.old_shape[f]:
                    change_shape()
                    return shape
                elif shape[f] < self.old_shape[f]:
                    if time.time() > (self.last_reshaped + self.reshape_delay):
                        change_shape()
                        return shape
        return self.old_shape

    def render_table(self, data, fields):
        prepared, shape = self.__prepare(data)
        headers = {}
        for f in shape:
            if 'header' in self.template[f]:
                headers[f] = self.template[f]['header']
                shape[f] = max(shape[f], len(headers[f]))
            else:
                headers[f] = ''
        shape = self.__update_shape(shape)
        has_headers = any(headers.values())
        delimiter_gen = self.__delimiter_gen()
        row_tpl = ''
        for num, field in enumerate(fields):
            if 'final' in self.template[field]:
                final = self.template[field]['final']
            else:
                final = self.default_final
            row_tpl += final.format(field, len=shape[field])
            if num < len(fields) - 1:
                row_tpl += next(delimiter_gen)
        result = []
        if has_headers:
            result.append(
                (row_tpl.format(**headers),))
        for line in prepared:
            result.append(
                (row_tpl.format(**line),))
        return result


class Sparkline(object):
    def __init__(self, window):
        self.log = logging.getLogger(__name__)
        self.data = {}
        self.window = window
        self.active_seconds = []
        self.ticks = '_▁▂▃▄▅▆▇'

    def recalc_active(self, ts):
        if not self.active_seconds:
            self.active_seconds.append(ts)
            self.data[ts] = {}
        if ts not in self.active_seconds:
            if ts > max(self.active_seconds):
                for i in range(max(self.active_seconds) + 1, ts + 1):
                    self.active_seconds.append(i)
                    self.active_seconds.sort()
                    self.data[i] = {}
        while len(self.active_seconds) > self.window:
            self.active_seconds.pop(0)
        for sec in tuple(self.data.keys()):
            if sec not in self.active_seconds:
                self.data.pop(sec)

    def get_key_data(self, key):
        result = []
        if not self.active_seconds:
            return None
        for sec in self.active_seconds:
            if key in self.data[sec]:
                result.append(self.data[sec][key])
            else:
                result.append(('', 0))
        return result

    def add(self, ts, key, value, color=''):
        if ts not in self.data:
            self.recalc_active(ts)
        if ts < min(self.active_seconds):
            self.log.warning('Sparkline got outdated second %s, oldest in list %s', ts, min(self.active_seconds))
            return
        value = max(value, 0)
        self.data[ts][key] = (color, value)

    def get_sparkline(self, key, baseline='zero', spark_len='auto', align='right'):
        if spark_len == 'auto':
            spark_len = self.window
        elif spark_len <= 0:
            return ''
        key_data = self.get_key_data(key)
        if not key_data:
            return ''
        active_data = key_data[-spark_len:]
        result = []
        if active_data:
            values = [i[1] for i in active_data]
            if baseline == 'zero':
                min_val = 0
                step = float(max(values)) / len(self.ticks)
            elif baseline == 'min':
                min_val = min(values)
                step = float(max(values) - min_val) / len(self.ticks)
            ranges = [step * i for i in range(len(self.ticks) + 1)]
            for color, value in active_data:
                if value <= 0:
                    tick = ' '
                else:
                    rank = bisect.bisect_left(ranges, value) - 1
                    rank = max(rank, 0)
                    rank = min(rank, len(self.ticks) - 1)
                    tick = self.ticks[rank]
                result.append(color)
                result.append(tick)
        space = ' ' * (spark_len - len(active_data))
        if align == 'right':
            result = [space] + result
        elif align == 'left':
            result = result + [space]
        return result


class Screen(object):
    '''     Console screen renderer class    '''
    RIGHT_PANEL_SEPARATOR = ' . '

    def __init__(self, info_panel_width, markup_provider, **kwargs):
        self.log = logging.getLogger(__name__)
        self.info_panel_percent = int(info_panel_width)
        self.info_widgets = {}
        self.markup = markup_provider
        self.term_height = 60
        self.term_width = 120
        self.right_panel_width = 10
        self.left_panel_width = self.term_width - self.right_panel_width - len(
            self.RIGHT_PANEL_SEPARATOR)
        cases_args = dict(
            [(k, v)
                for k, v in kwargs.items()
                if k in ['cases_sort_by', 'cases_max_spark', 'max_case_len']]
        )
        times_args = {'times_max_spark': kwargs['times_max_spark']}
        sizes_args = {'sizes_max_spark': kwargs['sizes_max_spark']}

        codes_block = VerticalBlock(CurrentHTTPBlock(self), CurrentNetBlock(self), self)
        times_block = VerticalBlock(AnswSizesBlock(self, **sizes_args), AvgTimesBlock(self, **times_args), self)
        top_block = VerticalBlock(codes_block, times_block, self)
        general_block = HorizontalBlock(PercentilesBlock(self), top_block, self)
        overall_block = VerticalBlock(RPSBlock(self), general_block, self)
        final_block = VerticalBlock(overall_block, CasesBlock(self, **cases_args), self)

        self.left_panel = final_block

    def __get_right_line(self, widget_output):
        '''        Gets next line for right panel        '''
        right_line = ''
        if widget_output:
            right_line = widget_output.pop(0)
            if len(right_line) > self.right_panel_width:
                right_line_plain = self.markup.clean_markup(right_line)
                if len(right_line_plain) > self.right_panel_width:
                    right_line = right_line[:self.right_panel_width] + self.markup.RESET
        return right_line

    def __truncate(self, line_arr, max_width):
        '''  Cut tuple of line chunks according to it's wisible lenght  '''
        def is_space(chunk):
            return all([True if i == ' ' else False for i in chunk])

        def is_empty(chunks, markups):
            result = []
            for chunk in chunks:
                if chunk in markups:
                    result.append(True)
                elif is_space(chunk):
                    result.append(True)
                else:
                    result.append(False)
            return all(result)
        left = max_width
        result = ''
        markups = self.markup.get_markup_vars()
        for num, chunk in enumerate(line_arr):
            if chunk in markups:
                result += chunk
            else:
                if left > 0:
                    if len(chunk) <= left:
                        result += chunk
                        left -= len(chunk)
                    else:
                        leftover = (chunk[left:],) + line_arr[num + 1:]
                        was_cut = not is_empty(leftover, markups)
                        if was_cut:
                            result += chunk[:left - 1] + self.markup.RESET + '…'
                        else:
                            result += chunk[:left]
                        left = 0
        return result

    def __render_left_panel(self):
        ''' Render left blocks '''
        self.log.debug("Rendering left blocks")
        left_block = self.left_panel
        left_block.render()
        blank_space = self.left_panel_width - left_block.width

        lines = []
        pre_space = ' ' * int(blank_space / 2)
        if not left_block.lines:
            lines = [(''), (self.markup.RED + 'BROKEN LEFT PANEL' + self.markup.RESET)]
        else:
            while self.left_panel.lines:
                src_line = self.left_panel.lines.pop(0)
                line = pre_space + self.__truncate(src_line, self.left_panel_width)
                post_space = ' ' * (self.left_panel_width - len(self.markup.clean_markup(line)))
                line += post_space + self.markup.RESET
                lines.append(line)
        return lines

    def render_screen(self):
        '''        Main method to render screen view        '''
        self.term_width, self.term_height = get_terminal_size()
        self.log.debug(
            "Terminal size: %sx%s", self.term_width, self.term_height)
        self.right_panel_width = int(
            (self.term_width - len(self.RIGHT_PANEL_SEPARATOR))
            * (float(self.info_panel_percent) / 100)) - 1
        if self.right_panel_width > 0:
            self.left_panel_width = self.term_width - \
                self.right_panel_width - len(self.RIGHT_PANEL_SEPARATOR) - 2
        else:
            self.right_panel_width = 0
            self.left_panel_width = self.term_width - 1
        self.log.debug(
            "Left/right panels width: %s/%s", self.left_panel_width,
            self.right_panel_width)

        widget_output = []
        if self.right_panel_width:
            widget_output = []
            self.log.debug("There are %d info widgets" % len(self.info_widgets))
            for index, widget in sorted(
                    self.info_widgets.items(),
                    key=lambda item: (item[1].get_index(), item[0])):
                self.log.debug("Rendering info widget #%s: %s", index, widget)
                widget_out = widget.render(self).strip()
                if widget_out:
                    widget_output += widget_out.split("\n")
                    widget_output += [""]

        left_lines = self.__render_left_panel()

        self.log.debug("Composing final screen output")
        output = []
        for line_no in range(1, self.term_height):
            line = " "

            if line_no > 1 and left_lines:
                left_line = left_lines.pop(0)
                left_line_plain = self.markup.clean_markup(left_line)

                left_line += (
                    ' ' * (self.left_panel_width - len(left_line_plain)))
                line += left_line
            else:
                line += ' ' * self.left_panel_width
            if self.right_panel_width:
                line += self.markup.RESET
                line += self.markup.WHITE
                line += self.RIGHT_PANEL_SEPARATOR
                line += self.markup.RESET
                right_line = self.__get_right_line(widget_output)
                line += right_line

            output.append(line)
        return self.markup.new_line.join(output) + self.markup.new_line

    def add_info_widget(self, widget):
        '''
        Add widget string to right panel of the screen
        '''
        index = widget.get_index()
        while index in self.info_widgets:
            index += 1
        self.info_widgets[widget.get_index()] = widget

    def add_second_data(self, data):
        '''
        Notification method about new aggregator data
        '''
        self.left_panel.add_second(data)


class AbstractBlock:
    '''
    Parent class for all left panel blocks
    '''

    def __init__(self, screen):
        self.log = logging.getLogger(__name__)
        self.lines = []
        self.width = 0
        self.screen = screen

    def add_second(self, data):
        '''
        Notification about new aggregate data
        '''
        pass

    def fill_rectangle(self, prepared):
        '''  Right-pad lines of block to equal width  '''
        result = []
        width = max([self.clean_len(line) for line in prepared])
        for line in prepared:
            spacer = ' ' * (width - self.clean_len(line))
            result.append(line + (self.screen.markup.RESET, spacer))
        return (width, result)

    def clean_len(self, line):
        '''  Calculate wisible length of string  '''
        if isinstance(line, str):
            return len(self.screen.markup.clean_markup(line))
        elif isinstance(line, tuple) or isinstance(line, list):
            markups = self.screen.markup.get_markup_vars()
            length = 0
            for i in line:
                if i not in markups:
                    length += len(i)
            return length

    def render(self):
        '''
        Render method, fills .lines and .width properties with rendered data
        '''
        raise RuntimeError("Abstract method needs to be overridden")


class HorizontalBlock(AbstractBlock):
    '''
    Block to merge two other blocks horizontaly
    '''

    def __init__(self, left_block, right_block, screen):
        AbstractBlock.__init__(self, screen)
        self.left = left_block
        self.right = right_block
        self.separator = '  . '

    def render(self, expected_width=None):
        if not expected_width:
            expected_width = self.screen.left_panel_width

        def get_line(source, num):
            if num >= len(source.lines):
                return (' ' * source.width,)
            else:
                source_line = source.lines[n]
                spacer = ' ' * (source.width - self.clean_len(source_line))
                return source_line + (spacer,)
        self.left.render(expected_width=expected_width)
        right_width_limit = expected_width - self.left.width - len(self.separator)
        self.right.render(expected_width=right_width_limit)
        self.height = max(len(self.left.lines), len(self.right.lines))
        self.width = self.left.width + self.right.width + len(self.separator)

        self.lines = []

        for n in range(self.height):
            self.lines.append(
                get_line(self.left, n) + (self.separator,) + get_line(self.right, n)
            )
        self.lines.append((' ' * self.width,))

    def add_second(self, data):
        self.left.add_second(data)
        self.right.add_second(data)


class VerticalBlock(AbstractBlock):
    '''
    Block to merge two other blocks vertically
    '''

    def __init__(self, top_block, bottom_block, screen):
        AbstractBlock.__init__(self, screen)
        self.top = top_block
        self.bottom = bottom_block

    def render(self, expected_width=None):
        if not expected_width:
            expected_width = self.screen.left_panel_width

        self.top.render(expected_width=expected_width)
        self.bottom.render(expected_width=expected_width)
        self.width = max(self.top.width, self.bottom.width)

        self.lines = []
        for line in self.top.lines:
            spacer = ' ' * (self.width - self.top.width)
            self.lines.append(line + (spacer,))

        if self.top.lines and self.bottom.lines:
            spacer = ' ' * self.width
            self.lines.append((spacer,))

        for line in self.bottom.lines:
            spacer = ' ' * (self.width - self.bottom.width)
            self.lines.append(line + (spacer,))

    def add_second(self, data):
        self.top.add_second(data)
        self.bottom.add_second(data)


class RPSBlock(AbstractBlock):
    ''' Actual RPS sparkline '''

    def __init__(self, screen):
        AbstractBlock.__init__(self, screen)
        self.begin_tpl = 'Data delay: {delay}s, RPS: {rps:>3,} '
        self.sparkline = Sparkline(180)
        self.last_count = 0
        self.last_second = None

    def add_second(self, data):
        count = data['overall']['interval_real']['len']
        self.last_count = count
        ts = data['ts']
        self.last_second = ts
        self.sparkline.add(ts, 'rps', count)

    def render(self, expected_width=None):
        if self.last_second:
            delay = int(time.time() - self.last_second)
        else:
            delay = ' - '
        line_start = self.begin_tpl.format(rps=self.last_count, delay=delay)
        spark_len = expected_width - len(line_start) - 2

        spark = self.sparkline.get_sparkline('rps', spark_len=spark_len)
        prepared = [(self.screen.markup.WHITE, line_start, ' ') + tuple(spark) + (self.screen.markup.RESET,)]
        self.width, self.lines = self.fill_rectangle(prepared)


class PercentilesBlock(AbstractBlock):
    ''' Aggregated percentiles '''

    def __init__(self, screen):
        AbstractBlock.__init__(self, screen)
        self.title = 'Percentiles (all/last 1m/last), ms:'
        self.overall = None
        self.width = 10
        self.last_ts = None
        self.last_min = {}
        self.quantiles = [10, 20, 30, 40, 50, 60, 70, 75, 80, 85, 90, 95, 99, 99.5, 100]
        template = {
            'quantile': {'tpl': '{:>.1f}%'},
            'all':      {'tpl': '{:>,.1f}'},  # noqa: E241
            'last_1m':  {'tpl': '{:>,.1f}'},  # noqa: E241
            'last':     {'tpl': '{:>,.1f}'}   # noqa: E241
        }
        delimiters = [' <  ', '  ']
        self.formatter = TableFormatter(template, delimiters)

    def add_second(self, data):
        incoming_hist = data['overall']['interval_real']['hist']
        ts = data['ts']
        self.precise_quantiles = {
            q: float(v) / 1000
            for q, v in zip(
                data["overall"]["interval_real"]["q"]["q"],
                data["overall"]["interval_real"]["q"]["value"])
        }
        dist = pd.Series(incoming_hist['data'], index=incoming_hist['bins'])
        if self.overall is None:
            self.overall = dist
        else:
            self.overall = self.overall.add(dist, fill_value=0)
        for second in tuple(self.last_min.keys()):
            if ts - second > 60:
                self.last_min.pop(second)
        self.last_min[ts] = dist

    def __calc_percentiles(self):
        def hist_to_quant(histogram, quant):
            cumulative = histogram.cumsum()
            total = cumulative.max()
            positions = cumulative.searchsorted([float(i) / 100 * total for i in quant])
            quant_times = [cumulative.index[i] / 1000. for i in positions]
            return quant_times

        all_times = hist_to_quant(self.overall, self.quantiles)
        last_data = self.last_min[max(self.last_min.keys())]
        last_times = hist_to_quant(last_data, self.quantiles)
        # Check if we have precise data for last second quantiles instead of binned histogram
        for position, q in enumerate(self.quantiles):
            if q in self.precise_quantiles:
                last_times[position] = self.precise_quantiles[q]
        # Replace binned values with precise, if lower quantile bin happens to be
        # greater than upper quantile precise values
        for position in reversed(range(1, len(last_times))):
            if last_times[position - 1] > last_times[position]:
                last_times[position - 1] = last_times[position]
        last_1m = pd.Series(dtype='int64')
        for ts, data in self.last_min.items():
            if last_1m.empty:
                last_1m = data
            else:
                last_1m = last_1m.add(data, fill_value=0)
        last_1m_times = hist_to_quant(last_1m, self.quantiles)
        quant_times = reversed(
            list(zip(self.quantiles, all_times, last_1m_times, last_times))
        )
        data = []
        for q, all_time, last_1m, last_time in quant_times:
            data.append({
                'quantile': q,
                'all': all_time,
                'last_1m': last_1m,
                'last': last_time
            })
        return data

    def render(self, expected_width=None):
        prepared = [(self.screen.markup.WHITE, self.title)]
        if self.overall is None:
            prepared.append(('',))
        else:
            data = self.__calc_percentiles()
            prepared += self.formatter.render_table(data, ['quantile', 'all', 'last_1m', 'last'])
        self.width, self.lines = self.fill_rectangle(prepared)


class CurrentHTTPBlock(AbstractBlock):
    ''' Http codes with highlight'''

    def __init__(self, screen):
        AbstractBlock.__init__(self, screen)
        self.overall_dist = defaultdict(int)
        self.title = 'HTTP codes: '
        self.total_count = 0
        template = {
            'count':       {'tpl': '{:>,}'},  # noqa: E241
            'last':        {'tpl': '+{:>,}'},  # noqa: E241
            'percent':     {'tpl': '{:>.2f}%'},  # noqa: E241
            'code':        {'tpl': '{:>3}', 'final': '{{{:}:<{len}}}'},  # noqa: E241
            'description': {'tpl': '{:<10}', 'final': '{{{:}:<{len}}}'}
        }
        delimiters = [' ', '  ', ' : ', ' ']
        self.formatter = TableFormatter(template, delimiters)

    def add_second(self, data):
        self.last_dist = data["overall"]["proto_code"]["count"]
        for code, count in self.last_dist.items():
            self.total_count += count
            self.overall_dist[code] += count

    def __code_color(self, code):
        colors = {(200, 299): self.screen.markup.GREEN,
                  (300, 399): self.screen.markup.CYAN,
                  (400, 499): self.screen.markup.YELLOW,
                  (500, 599): self.screen.markup.RED}
        if code in self.last_dist:
            for left, right in colors:
                if left <= int(code) <= right:
                    return colors[(left, right)]
            return self.screen.markup.MAGENTA
        else:
            return ''

    def __code_descr(self, code):
        if int(code) in util.HTTP:
            return util.HTTP[int(code)]
        else:
            return 'N/A'

    def render(self, expected_width=None):
        prepared = [(self.screen.markup.WHITE, self.title)]
        if not self.overall_dist:
            prepared.append(('',))
        else:
            data = []
            for code, count in sorted(self.overall_dist.items()):
                if code in self.last_dist:
                    last_count = self.last_dist[code]
                else:
                    last_count = 0
                data.append({
                    'count': count,
                    'last': last_count,
                    'percent': 100 * safe_div(count, self.total_count),
                    'code': code,
                    'description': self.__code_descr(code)
                })
            table = self.formatter.render_table(data, ['count', 'last', 'percent', 'code', 'description'])
            for num, line in enumerate(data):
                color = self.__code_color(line['code'])
                prepared.append((color, table[num][0]))
        self.width, self.lines = self.fill_rectangle(prepared)


class CurrentNetBlock(AbstractBlock):
    ''' NET codes with highlight'''

    def __init__(self, screen):
        AbstractBlock.__init__(self, screen)
        self.overall_dist = defaultdict(int)
        self.title = 'Net codes:'
        self.total_count = 0
        template = {
            'count':       {'tpl': '{:>,}'},  # noqa: E241
            'last':        {'tpl': '+{:>,}'},  # noqa: E241
            'percent':     {'tpl': '{:>.2f}%'},  # noqa: E241
            'code':        {'tpl': '{:>2}', 'final': '{{{:}:<{len}}}'},  # noqa: E241
            'description': {'tpl': '{:<10}', 'final': '{{{:}:<{len}}}'}
        }
        delimiters = [' ', '  ', ' : ', ' ']
        self.formatter = TableFormatter(template, delimiters)

    def add_second(self, data):
        self.last_dist = data["overall"]["net_code"]["count"]

        for code, count in self.last_dist.items():
            self.total_count += count
            self.overall_dist[code] += count

    def __code_descr(self, code):
        if int(code) in util.NET:
            return util.NET[int(code)]
        else:
            return 'N/A'

    def __code_color(self, code):
        if code in self.last_dist:
            if int(code) == 0:
                return self.screen.markup.GREEN
            elif int(code) == 314:
                return self.screen.markup.MAGENTA
            else:
                return self.screen.markup.RED
        else:
            return ''

    def render(self, expected_width=None):
        prepared = [(self.screen.markup.WHITE, self.title)]
        if not self.overall_dist:
            prepared.append(('',))
        else:
            data = []
            for code, count in sorted(self.overall_dist.items()):
                if code in self.last_dist:
                    last_count = self.last_dist[code]
                else:
                    last_count = 0
                data.append({
                    'count': count,
                    'last': last_count,
                    'percent': 100 * safe_div(count, self.total_count),
                    'code': code,
                    'description': self.__code_descr(code)
                })
            table = self.formatter.render_table(data, ['count', 'last', 'percent', 'code', 'description'])
            for num, line in enumerate(data):
                color = self.__code_color(line['code'])
                prepared.append((color, table[num][0]))
        self.width, self.lines = self.fill_rectangle(prepared)


class AnswSizesBlock(AbstractBlock):
    ''' Answer and response sizes, if available '''

    def __init__(self, screen, sizes_max_spark=120):
        AbstractBlock.__init__(self, screen)
        self.sparkline = Sparkline(sizes_max_spark)
        self.overall = {'count': 0, 'Response': 0, 'Request': 0}
        self.last = {'count': 0, 'Response': 0, 'Request': 0}
        self.title = 'Average Sizes (all/last), bytes:'
        template = {
            'name': {'tpl': '{:>}'},
            'avg': {'tpl': '{:>,.1f}'},
            'last_avg': {'tpl': '{:>,.1f}'}
        }
        delimiters = [': ', ' / ']
        self.formatter = TableFormatter(template, delimiters)

    def add_second(self, data):
        self.last['count'] = data["overall"]["interval_real"]["len"]
        self.last['Request'] = data["overall"]["size_out"]["total"]
        self.last['Response'] = data["overall"]["size_in"]["total"]

        self.overall['count'] += self.last['count']
        self.overall['Request'] += self.last['Request']
        self.overall['Response'] += self.last['Response']
        ts = data['ts']
        for direction in ['Request', 'Response']:
            self.sparkline.add(ts, direction, self.last[direction] / self.last['count'])

    def render(self, expected_width=None):
        prepared = [(self.screen.markup.WHITE, self.title)]
        if self.overall['count']:
            overall_avg = avg_from_dict(self.overall)
            last_avg = avg_from_dict(self.last)
            data = []
            for direction in ['Request', 'Response']:
                data.append({
                    'name': direction,
                    'avg': overall_avg[direction],
                    'last_avg': last_avg[direction]
                })
            table = self.formatter.render_table(data, ['name', 'avg', 'last_avg'])
            for num, direction in enumerate(['Request', 'Response']):
                spark_len = expected_width - self.clean_len(table[0]) - 3
                spark = self.sparkline.get_sparkline(direction, spark_len=spark_len)
                prepared.append(table[num] + ('  ',) + tuple(spark))
        else:
            self.lines.append(('',))
            self.lines.append(('',))
        self.width, self.lines = self.fill_rectangle(prepared)


class AvgTimesBlock(AbstractBlock):
    ''' Average times breakdown '''

    def __init__(self, screen, times_max_spark=120):
        AbstractBlock.__init__(self, screen)
        self.sparkline = Sparkline(times_max_spark)
        self.fraction_keys = [
            'interval_real', 'connect_time', 'send_time', 'latency', 'receive_time']
        self.fraction_names = {
            'interval_real': 'Overall',
            'connect_time':  'Connect',  # noqa: E241
            'send_time':     'Send',  # noqa: E241
            'latency':       'Latency',  # noqa: E241
            'receive_time':  'Receive'}  # noqa: E241
        self.overall = dict([(k, 0) for k in self.fraction_keys])
        self.overall['count'] = 0
        self.last = dict([(k, 0) for k in self.fraction_keys])
        self.last['count'] = 0
        self.title = 'Average Times (all/last), ms:'
        template = {
            'name': {'tpl': '{:>}'},
            'avg': {'tpl': '{:>,.2f}'},
            'last_avg': {'tpl': '{:>,.2f}'}
        }
        delimiters = [': ', ' / ']
        self.formatter = TableFormatter(template, delimiters)

    def add_second(self, data):
        self.last = {}
        self.last['count'] = data["overall"]["interval_real"]["len"]
        self.overall['count'] += self.last['count']
        ts = data["ts"]
        for fraction in self.fraction_keys:
            self.last[fraction] = float(data["overall"][fraction]["total"]) / 1000
            self.overall[fraction] += self.last[fraction]
            self.sparkline.add(ts, fraction, self.last[fraction] / self.last['count'])

    def render(self, expected_width=None):
        prepared = [(self.screen.markup.WHITE, self.title)]
        if self.overall['count']:
            overall_avg = avg_from_dict(self.overall)
            last_avg = avg_from_dict(self.last)
            data = []
            for fraction in self.fraction_keys:
                data.append({
                    'name': self.fraction_names[fraction],
                    'avg': overall_avg[fraction],
                    'last_avg': last_avg[fraction]
                })
            table = self.formatter.render_table(data, ['name', 'avg', 'last_avg'])
            for num, fraction in enumerate(self.fraction_keys):
                spark_len = expected_width - self.clean_len(table[0]) - 3
                spark = self.sparkline.get_sparkline(fraction, spark_len=spark_len)
                prepared.append(table[num] + ('  ',) + tuple(spark))
        else:
            for fraction in self.fraction_keys:
                prepared.append(('-',))
        self.width, self.lines = self.fill_rectangle(prepared)


class CasesBlock(AbstractBlock):
    '''     Cases info    '''

    def __init__(self, screen, cases_sort_by='http_err', cases_max_spark=60, reorder_delay=5, max_case_len=32):
        AbstractBlock.__init__(self, screen)
        self.cumulative_cases = {}
        self.last_cases = {}
        self.title = 'Cumulative Cases Info:'
        self.max_case_len = max_case_len
        self.cases_order = []
        self.reorder_delay = reorder_delay
        self.sparkline = Sparkline(cases_max_spark)
        self.cases_sort_by = cases_sort_by
        self.last_reordered = time.time()
        self.field_order = ['name', 'count', 'percent', 'last', 'net_err', 'http_err', 'avg', 'last_avg']

        template = {
            'name':     {'tpl': '{:>}:',    'header': 'name', 'final': '{{{:}:>{len}}}'},  # noqa: E241
            'count':    {'tpl': '{:>,}',    'header': 'count'},   # noqa: E241
            'last':     {'tpl': '+{:>,}',   'header': 'last'},    # noqa: E241
            'percent':  {'tpl': '{:>.2f}%', 'header': '%'},       # noqa: E241
            'net_err':  {'tpl': '{:>,}',    'header': 'net_e'},   # noqa: E241
            'http_err': {'tpl': '{:>,}',    'header': 'http_e'},  # noqa: E241
            'avg':      {'tpl': '{:>,.1f}', 'header': 'avg ms'},  # noqa: E241
            'last_avg': {'tpl': '{:>,.1f}', 'header': 'last ms'}
        }
        delimiters = [' ']
        self.formatter = TableFormatter(template, delimiters)

    def add_second(self, data):
        def prepare_data(tag_data, display_name):
            count = tag_data["interval_real"]["len"]
            time = tag_data["interval_real"]["total"] / 1000
            net_err, http_err, spark_color = combine_codes(tag_data, self.screen.markup)
            if spark_color == self.screen.markup.GREEN:
                text_color = self.screen.markup.WHITE
            else:
                text_color = spark_color
            return (spark_color, {
                'count': count, 'net_err': net_err, 'http_err': http_err,
                'time': time, 'color': text_color, 'display_name': display_name})

        ts = data["ts"]
        overall = data["overall"]
        self.last_cases = {}
        spark_color, self.last_cases[0] = prepare_data(overall, 'OVERALL')
        self.sparkline.add(ts, 0, self.last_cases[0]['count'], color=spark_color)

        tagged = data["tagged"]
        for tag_name, tag_data in tagged.items():
            # decode symbols to utf-8 in order to support cyrillic symbols in cases
            # name = tag_name.decode('utf-8')
            spark_color, self.last_cases[tag_name] = prepare_data(tag_data, tag_name)
            self.sparkline.add(ts, tag_name, self.last_cases[tag_name]['count'], color=spark_color)
        for name in self.last_cases:
            if name not in self.cumulative_cases:
                self.cumulative_cases[name] = {}
                for k in ['count', 'net_err', 'http_err', 'time', 'display_name']:
                    self.cumulative_cases[name][k] = self.last_cases[name][k]
            else:
                for k in ['count', 'net_err', 'http_err', 'time']:
                    self.cumulative_cases[name][k] += self.last_cases[name][k]

    def __cut_name(self, name):
        if len(name) > self.max_case_len:
            return name[:self.max_case_len - 1] + '…'
        else:
            return name

    def __reorder_cases(self):
        sorted_cases = sorted(self.cumulative_cases.items(),
                              key=lambda item: (-1 * item[1][self.cases_sort_by], str(item[0])))
        new_order = [case for (case, data) in sorted_cases]
        now = time.time()
        if now - self.reorder_delay > self.last_reordered:
            self.cases_order = new_order
            self.last_reordered = now
        else:
            if len(new_order) > len(self.cases_order):
                for case in new_order:
                    if case not in self.cases_order:
                        self.cases_order.append(case)

    def render(self, expected_width=None):
        prepared = [(self.screen.markup.WHITE, self.title)]

        if 0 in self.cumulative_cases:  # 0 used as special name for OVERALL to avoid name collision
            total_count = self.cumulative_cases[0]['count']

            self.__reorder_cases()
            data = []
            for name in self.cases_order:
                case_data = self.cumulative_cases[name]
                if name in self.last_cases:
                    last = self.last_cases[name]
                else:
                    last = {'count': 0, 'net_err': 0, 'http_err': 0, 'time': 0, 'color': '', 'display_name': name}
                data.append({
                    'full_name': name,
                    'name': self.__cut_name(case_data['display_name']),
                    'count': case_data['count'],
                    'percent': 100 * safe_div(case_data['count'], total_count),
                    'last': last['count'],
                    'net_err': case_data['net_err'],
                    'http_err': case_data['http_err'],
                    'avg': safe_div(case_data['time'], case_data['count']),
                    'last_avg': safe_div(last['time'], last['count'])
                })
            table = self.formatter.render_table(data, self.field_order)
            prepared.append(table[0])  # first line is table header
            for num, line in enumerate(data):
                full_name = line['full_name']
                if full_name in self.last_cases:
                    color = self.last_cases[full_name]['color']
                else:
                    color = ''
                spark_len = expected_width - self.clean_len(table[0]) - 3
                spark = self.sparkline.get_sparkline(full_name, spark_len=spark_len)
                prepared.append((color,) + table[num + 1] + ('  ',) + tuple(spark))

        for _ in range(3 - len(self.cumulative_cases)):
            prepared.append(('',))

        self.width, self.lines = self.fill_rectangle(prepared)



"""Known metrics decoder"""
import logging

logger = logging.getLogger(__name__)


class MetricsDecoder(object):
    def __init__(self):
        """
        translates telegraf metric names into common Monitoring metric names
        translates `uncommon` names to `custom:%s`s

        """
        self.known_metrics = {
            'mem_used': 'Memory_used',
            'mem_free': 'Memory_free',
            'mem_buffered': 'Memory_buff',
            'mem_cached': 'Memory_cached',
            'kernel_context_switches': 'System_csw',
            'kernel_interrupts': 'System_int',
            'kernel_processes_forked': 'System_forks',
            'processes_total': 'System_numproc',
            'processes_total_threads': 'System_numthreads',
            'system_load1': 'System_la1',
            'system_load5': 'System_la5',
            'system_load15': 'System_la15',
            'nstat_TcpRetransSegs': 'Net_retransmit',
            # those guys became inactive due to net interface names and disk ids
            # we don't need unknown id data here
            # 'net_packets_recv': 'Net_rx',
            # 'net_packets_sent': 'Net_tx',
            # 'net_bytes_recv': 'Net_recv',
            # 'net_bytes_sent': 'Net_send',
            # 'diskio_read_bytes': 'Disk_read',
            # 'diskio_write_bytes': 'Disk_write',
            # ----------
            # remove this crunch after front refactoring
            # 'cpu-cpu-total_usage_user': 'CPU_user',
            # 'cpu-cpu-total_usage_system': 'CPU_system',
            # 'cpu-cpu-total_usage_idle': 'CPU_idle',
            # 'cpu-cpu-total_usage_iowait': 'CPU_iowait',
            # 'cpu-cpu-total_usage_irq': 'CPU_irq',
            # 'cpu-cpu-total_usage_nice': 'CPU_nice',
            # 'cpu-cpu-total_usage_softirq': 'CPU_softirq',
            # 'cpu-cpu-total_usage_steal': 'CPU_steal',
            # 'cpu-cpu-total_usage_guest': 'CPU_guest'
        }

        self.diff_metrics = {
            'cpu': [],
            'mem': [],
            'net': ['packets_recv', 'packets_sent', 'bytes_recv', 'bytes_sent'],
            'nstat': ['TcpRetransSegs'],
            'net_response': [],
            'kernel': ['context_switches', 'interrupts', 'processes_forked', 'vmstat_pgfault', 'vmstat_pgmajfault'],
            'diskio': [
                'read_bytes', 'write_bytes', 'io_time', 'read_time', 'reads',
                'write_time', 'writes'
            ],
            'custom': []
        }

    def find_common_names(self, key):
        if key in self.known_metrics:
            return self.known_metrics[key]
        else:
            return 'custom:{}'.format(key)


decoder = MetricsDecoder()



'''
Load Plan generators
'''
import re
from itertools import chain, groupby
from builtins import range
from . import info
from .util import parse_duration, solve_quadratic, proper_round


class Const(object):
    '''
    Load plan with constant load
    '''

    def __init__(self, rps, duration):
        self.rps = float(rps)
        self.duration = duration

    def __iter__(self):
        if self.rps == 0:
            return iter([])
        interval = 1000.0 / self.rps
        return (
            int(i * interval)
            for i in range(0, int(self.rps * self.duration / 1000)))

    def rps_at(self, t):
        '''Return rps for second t'''
        if 0 <= t <= self.duration:
            return self.rps
        else:
            return 0

    def get_duration(self):
        '''Return step duration'''
        return self.duration

    def __len__(self):
        '''Return total ammo count'''
        return int(self.duration / 1000 * self.rps)

    def get_rps_list(self):
        return [(int(self.rps), self.duration / 1000)]

    def __repr__(self):
        return 'const(%s, %s)' % (self.rps, self.duration / 1000)


class Line(object):
    '''Load plan with linear load'''

    def __init__(self, minrps, maxrps, duration):
        """

        :param minrps:
        :param maxrps:
        :param duration: milliseconds
        """
        self.minrps = float(minrps)
        self.maxrps = float(maxrps)
        self.duration = duration / 1000.0
        self.slope = (self.maxrps - self.minrps) / self.duration

    def ts(self, n):
        """
        :param n: number of charge
        :return: when to shoot nth charge, milliseconds
        """
        try:
            root1, root2 = solve_quadratic(self.slope / 2.0, self.minrps, -n)
        except ZeroDivisionError:
            root2 = float(n) / self.minrps
        return int(root2 * 1000)

    def __iter__(self):
        """

        :return: timestamps for each charge
        """
        return (self.ts(n) for n in range(0, self.__len__()))

    def rps_at(self, t):
        '''Return rps for second t'''
        if 0 <= t <= self.duration:
            return self.minrps + \
                float(self.maxrps - self.minrps) * t / self.duration
        else:
            return 0

    def get_duration(self):
        '''Return load duration in seconds'''
        return int(self.duration * 1000)

    def __len__(self):
        '''Return total ammo count'''
        return int((self.maxrps + self.minrps) / 2.0 * self.duration)

    def get_float_rps_list(self):
        '''
        get list of constant load parts (we have no constant load at all, but tank will think so),
        with parts durations (float)
        '''
        int_rps = range(int(self.minrps), int(self.maxrps) + 1)
        step_duration = float(self.duration) / len(int_rps)
        rps_list = [(rps, int(step_duration)) for rps in int_rps]
        return rps_list

    def get_rps_list(self):
        """
        get list of each second's rps
        :returns: list of tuples (rps, duration of corresponding rps in seconds)
        :rtype: list
        """
        seconds = range(0, int(self.duration) + 1)
        rps_groups = groupby([proper_round(self.rps_at(t)) for t in seconds],
                             lambda x: x)
        rps_list = [(rps, len(list(rpl))) for rps, rpl in rps_groups]
        return rps_list


class Composite(object):
    '''Load plan with multiple steps'''

    def __init__(self, steps):
        self.steps = steps

    def __iter__(self):
        base = 0
        for step in self.steps:
            for ts in step:
                yield ts + base
            base += step.get_duration()

    def get_duration(self):
        '''Return total duration'''
        return sum(step.get_duration() for step in self.steps)

    def __len__(self):
        '''Return total ammo count'''
        return int(sum(step.__len__() for step in self.steps))

    def get_rps_list(self):
        return list(
            chain.from_iterable(step.get_rps_list() for step in self.steps))


class Stairway(Composite):
    def __init__(self, minrps, maxrps, increment, step_duration):
        self.duration = step_duration
        if maxrps < minrps:
            increment = -increment
        n_steps = int((maxrps - minrps) / increment)
        steps = [
            Const(minrps + i * increment, step_duration)
            for i in range(0, n_steps + 1)
        ]
        if increment > 0:
            if (minrps + n_steps * increment) < maxrps:
                steps.append(Const(maxrps, step_duration))
        elif increment < 0:
            if (minrps + n_steps * increment) > maxrps:
                steps.append(Const(maxrps, step_duration))
        super(Stairway, self).__init__(steps)


class StepFactory(object):
    DURATION_RE = r'([0-9.]+d)?([0-9.]+h)?([0-9.]+m)?([0-9.]+s?)?'

    @classmethod
    def line(cls, params):
        template = re.compile(r'([0-9.]+),\s*([0-9.]+),\s*({})\)'.format(cls.DURATION_RE))
        minrps, maxrps, duration = template.search(params).groups()[:3]
        return Line(float(minrps), float(maxrps), parse_duration(duration))

    @classmethod
    def const(cls, params):
        template = re.compile(r'([0-9.]+),\s*({})\)'.format(cls.DURATION_RE))
        rps, duration = template.search(params).groups()[:2]
        return Const(float(rps), parse_duration(duration))

    @classmethod
    def stairway(cls, params):
        template = re.compile(
            r'([0-9.]+),\s*([0-9.]+),\s*([0-9.]+),\s*({})\)'.format(cls.DURATION_RE))
        minrps, maxrps, increment, duration = template.search(params).groups()[:4]
        return Stairway(
            float(minrps),
            float(maxrps), float(increment), parse_duration(duration))

    @staticmethod
    def produce(step_config):
        _plans = {
            'line': StepFactory.line,
            'const': StepFactory.const,
            'step': StepFactory.stairway,
        }
        load_type, params = step_config.split('(')
        load_type = load_type.strip()
        if load_type in _plans:
            return _plans[load_type](params)
        else:
            raise NotImplementedError('No such load type implemented: "%s"' % load_type)


def create(rps_schedule):
    """
    Create Load Plan as defined in schedule. Publish info about its duration.
    """
    if len(rps_schedule) > 1:
        lp = Composite(
            [StepFactory.produce(step_config) for step_config in rps_schedule])
    else:
        lp = StepFactory.produce(rps_schedule[0])
    info.status.publish('duration', lp.get_duration() / 1000)
    info.status.publish('steps', lp.get_rps_list())
    info.status.lp_len = len(lp)
    return lp



import os
from queue import Queue

import logging
from yandextank.common.interfaces import MonitoringPlugin
from yandextank.common.monitoring import MonitoringPanel, DefaultCollector
from yandextank.common.util import expand_to_seconds
from yandextank.plugins.YCMonitoring.sensor import YCMonitoringSensor

LOGGER = logging.getLogger(__name__)


def as_list(value) -> list | None:
    if value is None or isinstance(value, list):
        return value
    return [value]


class Plugin(MonitoringPlugin):
    def __init__(self, core, cfg, name):
        super(Plugin, self).__init__(core, cfg, name)
        self.timeout = expand_to_seconds(self.get_option('timeout'))
        self.poll_interval = expand_to_seconds(self.get_option('poll_interval'))
        self.request_timeout = expand_to_seconds(self.get_option('request_timeout'))

    def prepare_test(self):
        token = None
        token_option = self.get_option('token')
        if token_option == 'LOADTESTING_YC_TOKEN':
            token = os.environ.get('LOADTESTING_YC_TOKEN')
        else:
            try:
                with open(token_option, 'r') as tfile:
                    token = tfile.read().strip('\n')
            except (OSError, IOError):
                error = f"YCMonitoring plugin: Authorization token is not set! File {self.get_option('token')} is not found or can't be read."
                LOGGER.warning(error, exc_info=True)

        if token:
            self.collector = DefaultCollector(logger=LOGGER, timeout=self.timeout, poll_interval=self.poll_interval)
            api_host = self.get_option('api_host')
            for panel_name, query_data in self.get_option('panels').items():
                queue = Queue()
                group_name = query_data.get('group_name') or api_host
                panel = MonitoringPanel(group_name, self.timeout, queue)
                senset = set()
                for query in query_data.get('queries'):
                    sensor = YCMonitoringSensor(
                        panel_name,
                        api_host,
                        token,
                        query,
                        query_data.get('folder_id') or os.environ.get('LOADTESTING_FOLDER_ID'),
                        panel.queue,
                        self.request_timeout,
                        as_list(self.get_option('priority_labels')),
                        as_list(self.get_option('ignore_labels')),
                    )
                    try:
                        senset.update(sensor.get_sensors())
                        self.collector.add_sensor(sensor)
                    except ConnectionError:
                        LOGGER.warning(f'ConnectionError when trying to get sensors with query: {query}', exc_info=True)
                panel.add_sensors(senset)
                self.collector.add_panel(panel)
        else:
            LOGGER.warning('YC Token not found')



# -*- coding: UTF-8 -*-
import logging
import numpy as np
import time
from collections import Counter

logger = logging.getLogger(__name__)

phout_columns = [
    'time', 'tag', 'interval_real', 'connect_time', 'send_time', 'latency',
    'receive_time', 'interval_event', 'size_out', 'size_in', 'net_code',
    'proto_code'
]

phantom_config = {
    "interval_real": ["total", "max", "min", "hist", "q", "len"],
    "connect_time": ["total", "max", "min", "len"],
    "send_time": ["total", "max", "min", "len"],
    "latency": ["total", "max", "min", "len"],
    "receive_time": ["total", "max", "min", "len"],
    "interval_event": ["total", "max", "min", "len"],
    "size_out": ["total", "max", "min", "len"],
    "size_in": ["total", "max", "min", "len"],
    "net_code": ["count"],
    "proto_code": ["count"],
}


class Worker(object):
    """
    Aggregate Pandas dataframe or dict with numpy ndarrays in it
    """

    def __init__(self, config, verbose_histogram):
        if verbose_histogram:
            bins = np.linspace(0, 4990, 500)  # 10µs accuracy
            bins = np.append(bins,
                             np.linspace(5000, 9900, 50))  # 100µs accuracy
            bins = np.append(bins,
                             np.linspace(10, 499, 490) * 1000)  # 1ms accuracy
            bins = np.append(bins,
                             np.linspace(500, 2995, 500) * 1000)  # 5ms accuracy
            bins = np.append(bins, np.linspace(3000, 9990, 700) * 1000)  # 10ms accuracy
            bins = np.append(bins, np.linspace(10000, 29950, 400) * 1000)  # 50ms accuracy
            bins = np.append(bins, np.linspace(30000, 119900, 900) * 1000)  # 100ms accuracy
            bins = np.append(bins, np.linspace(120, 300, 181) * 1000000)  # 1s accuracy
        else:
            # yapf: disable
            bins = np.array([
                0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
                10, 20, 30, 40, 50, 60, 70, 80, 90,
                100, 150, 200, 250, 300, 350, 400, 450,
                500, 600, 650, 700, 750, 800, 850, 900, 950,
                1000, 1500, 2000, 2500, 3000, 3500, 4000, 4500,
                5000, 5500, 6000, 6500, 7000, 7500, 8000, 8500, 9000, 9500, 10000, 11000,
                12000, 13000, 14000, 15000, 20000, 25000, 30000, 35000, 40000, 45000, 50000,
                55000, 60000,
            ]) * 1000
            # yapf: enable

        self.bins = bins
        self.percentiles = np.array([50, 75, 80, 85, 90, 95, 98, 99, 100])
        self.config = config
        self.aggregators = {
            "hist": self._histogram,
            "q": self._quantiles,
            "mean": self._mean,
            "total": self._total,
            "min": self._min,
            "max": self._max,
            "count": self._count,
            "len": self._len,
        }

    def _histogram(self, series):
        data, bins = np.histogram(series, bins=self.bins)
        mask = data > 0
        return {
            "data": [e.item() for e in data[mask]],
            "bins": [e.item() for e in bins[1:][mask]],
        }

    def _mean(self, series):
        return series.mean().item()

    def _total(self, series):
        return series.sum().item()

    def _max(self, series):
        return series.max().item()

    def _min(self, series):
        return series.min().item()

    def _count(self, series):
        return {str(k): v for k, v in dict(Counter(series)).items()}

    def _len(self, series):
        return len(series)

    def _quantiles(self, series):
        return {
            "q": list(self.percentiles),
            "value": list(np.percentile(series, self.percentiles)),
        }

    def aggregate(self, data):
        return {
            key: {
                aggregate: self.aggregators.get(aggregate)(data[key])
                for aggregate in self.config[key]
            }
            for key in self.config
        }


class DataPoller:
    def __init__(self, *, poll_period=0.5, max_wait):
        self._pool_period = poll_period
        self._max_wait = max_wait

    def poll(self, source):
        return _data_poller(
            source,
            self._pool_period,
            self._max_wait
        )


def _data_poller(source, poll_period, max_wait):
    wait_cntr_max = max_wait // poll_period or 1
    wait_counter = 0
    for chunk in source:
        if chunk is not None:
            wait_counter = 0
            yield chunk
        elif wait_counter < wait_cntr_max:
            wait_counter += 1
        else:
            logger.warning('Data poller has been receiving no data for {} seconds.\n'
                           'Closing data poller'.format(wait_cntr_max * poll_period))
            break
        time.sleep(poll_period)


def to_utc(ts):
    # dst = daylight saving time
    is_dst = time.daylight and time.localtime().tm_isdst > 0
    offset = (time.altzone if is_dst else time.timezone)
    import pdb
    pdb.set_trace()
    return ts + offset


class Aggregator(object):
    def __init__(self, source, config, verbose_histogram=True):
        self.worker = Worker(config, verbose_histogram)
        self.source = source
        self.groupby = 'tag'

    def __iter__(self):
        for ts, chunk, rps in self.source:
            by_tag = list(chunk.groupby(self.groupby))
            start_time = time.time()
            result = {
                "ts": ts,
                "tagged":
                {tag: self.worker.aggregate(data)
                 for tag, data in by_tag},
                "overall": self.worker.aggregate(chunk),
                "counted_rps": rps
            }
            logger.debug(
                "Aggregation time: %.2fms", (time.time() - start_time) * 1000)
            yield result



import argparse
import glob
import json
import logging
import os
import socket

import sys

import pwd
import threading
from urllib.parse import urljoin

from datetime import datetime
import pkg_resources
import yaml
from cerberus import Validator

from yandextank.core import TankCore
from yandextank.core.tankcore import VALIDATED_CONF
from yandextank.validator.validator import ValidationError, load_yaml_schema
from .client import APIClient, OverloadClient, LPRequisites
from .plugin import LPJob, BackendTypes
from .plugin import Plugin as DataUploader

DATA_LOG = 'test_data.log'
MONITORING_LOG = 'monitoring.log'
SECTION = 'meta'


def get_logger():
    # TODO: remove global
    global logger   # pylint: disable=global-variable-undefined
    logger = logging.getLogger('')
    logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(stream=sys.stdout)
    handler.setLevel(logging.INFO)
    logger.addHandler(handler)
    verbose_handler = logging.FileHandler(
        datetime.now().strftime("post_loader_%Y-%m-%d_%H-%M-%S.log"), 'w')
    verbose_handler.setLevel(logging.DEBUG)
    logger.addHandler(verbose_handler)


def from_tank_config(test_dir):
    try:
        config_file = glob.glob(os.path.join(test_dir, VALIDATED_CONF))[0]
        logger.info('Config file found: %s' % config_file)
    except IndexError:
        raise OSError('Config file {} not found in {}'.format(VALIDATED_CONF, test_dir))

    with open(config_file) as f:
        tank_cfg = yaml.load(f, Loader=yaml.FullLoader)
    try:
        section, config = next(filter(
            lambda item: 'DataUploader' in item[1].get('package', ''),
            tank_cfg.items(),
        ))
    except StopIteration:
        logger.warning('DataUploader configuration not found in {}'.format(config_file))
        section, config = None, {}
    return section, config


def check_log(log_name):
    assert os.path.exists(log_name), \
        'Data log {} not found\n'.format(log_name) + \
        'JsonReport plugin should be enabled when launching Yandex-tank'


def upload_data(shooting_dir, log_name, lp_job):
    data_log = os.path.join(shooting_dir, log_name)
    check_log(data_log)
    sys.stdout.write('Uploading test data')
    with open(data_log, 'r') as f:
        for line in f:
            data = json.loads(line.strip())
            lp_job.push_test_data(data['data'], data['stats'])
            sys.stdout.write('.')
            sys.stdout.flush()
    sys.stdout.write('\n')


def upload_monitoring(shooting_dir, log_name, lp_job):
    data_log = os.path.join(shooting_dir, log_name)
    check_log(data_log)
    sys.stdout.write('Uploading monitoring data')
    with open(data_log, 'r') as f:
        for line in f.readlines():
            lp_job.push_monitoring_data(json.loads(line.strip()))
            sys.stdout.write('.')
            sys.stdout.flush()
    sys.stdout.write('\n')


def send_config_snapshot(config, lp_job):
    lp_job.send_config(LPRequisites.CONFIGINFO, yaml.dump(config))


def edit_metainfo(lp_config, lp_job):
    lp_job.edit_metainfo(regression_component=lp_config.get('component'),
                         cmdline=lp_config.get('cmdline'),
                         ammo_path=lp_config.get('ammo_path'),
                         loop_count=lp_config.get('loop_count'))


def get_plugin_dir(shooting_dir):
    DIRNAME = 'lunapark'
    parent = os.path.abspath(os.path.join(shooting_dir, os.pardir))
    if os.path.basename(parent) == DIRNAME:
        return parent
    else:
        plugin_dir = os.path.join(parent, DIRNAME)
        if not os.path.exists(plugin_dir):
            os.makedirs(plugin_dir)
        return plugin_dir


def make_symlink(shooting_dir, name):
    plugin_dir = get_plugin_dir(shooting_dir)
    link_name = os.path.join(plugin_dir, str(name))
    try:
        os.symlink(os.path.relpath(shooting_dir, plugin_dir), link_name)
    except OSError:
        logger.warning('Unable to create symlink for artifact: %s', link_name)
    else:
        logger.info('Symlink created: {}'.format(os.path.abspath(link_name)))


class ConfigError(Exception):
    pass


def post_loader():
    CONFIG_SCHEMA = load_yaml_schema(pkg_resources.resource_filename('yandextank.plugins.DataUploader',
                                                                     'config/postloader_schema.yaml'))

    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-c', '--config', help='YAML config. Format:\n{}'.format(yaml.dump(CONFIG_SCHEMA)))
    parser.add_argument('-a', '--api_address',
                        help='service to upload test results to, e.g. https://overload.yandex.net')
    parser.add_argument('-t', '--target', help='Address of the tested target, host[:port]')
    parser.add_argument('-o', '--operator', help='User who conducted the test')
    parser.add_argument('--task', help='task name, for Lunapark users only')
    parser.add_argument('--job_name', help='Job name')
    parser.add_argument('--job_dsc', help='Job description')
    parser.add_argument('--token', help='path to token file, for Overload users only')
    parser.add_argument('test_dir',
                        help='Directory containing test artifacts')
    args = parser.parse_args()
    assert os.path.exists(args.test_dir), 'Directory {} not found'.format(args.test_dir)
    get_logger()
    # load cfg
    if args.config:
        with open(args.config) as f:
            config = yaml.load(f, Loader=yaml.FullLoader)
            section = None
    else:
        section, config = from_tank_config(args.test_dir)
    # parse target host and port
    if args.target is not None:
        try:
            target_host, target_port = args.target.rsplit(':', 1)
        except ValueError:
            target_host, target_port = args.target, None
    else:
        target_host, target_port = None, None
    # update cfg from cli options
    for key, value in [('api_address', args.api_address),
                       ('target_host', target_host),
                       ('target_port', target_port),
                       ('operator', args.operator),
                       ('task', args.task),
                       ('job_name', args.job_name),
                       ('job_dsc', args.job_dsc),
                       ('token_file', args.token)]:
        if value is not None:
            config[key] = value
    # Validation
    v = Validator(schema=CONFIG_SCHEMA,
                  allow_unknown=True)
    if not v.validate(config):
        raise ValidationError(v.errors)
    config = v.normalized(config)

    # lunapark or overload?
    backend_type = BackendTypes.identify_backend(config['api_address'], section)
    if backend_type == BackendTypes.LUNAPARK:
        client = APIClient
        api_token = None
    elif backend_type == BackendTypes.OVERLOAD:
        client = OverloadClient
        try:
            api_token = DataUploader.read_token(config["token_file"])
        except KeyError:
            raise ConfigError('Token file required')
    else:
        raise RuntimeError("Backend type doesn't match any of the expected")

    user_agent = ' '.join(('Uploader/{}'.format(DataUploader.VERSION),
                           TankCore.get_user_agent()))
    api_client = client(base_url=config['api_address'],
                        user_agent=user_agent,
                        api_token=api_token,
                        core_interrupted=threading.Event()
                        # todo: add timeouts
                        )
    lp_job = LPJob(
        client=api_client,
        target_host=config.get('target_host'),
        target_port=config.get('target_port'),
        person=config.get('operator') or pwd.getpwuid(os.geteuid())[0],
        task=config.get('task'),
        name=config['job_name'],
        description=config['job_dsc'],
        tank=socket.getfqdn())
    edit_metainfo(config, lp_job)
    upload_data(args.test_dir, DATA_LOG, lp_job)
    send_config_snapshot(config, lp_job)
    try:
        upload_monitoring(args.test_dir, MONITORING_LOG, lp_job)
    except AssertionError as e:
        logger.error(e)
    lp_job.close(0)
    make_symlink(args.test_dir, lp_job.number)
    logger.info(
        'LP job created: {}'.format(
            urljoin(
                api_client.base_url, str(
                    lp_job.number))))


if __name__ == '__main__':
    post_loader()



"""
Global metrics publishing module. Inspired by Golang's expvar module

This implementation is not thread-safe
"""

from queue import Queue, Empty
import time


class ExpVar(object):
    """
    This class stores variables
    """

    def __init__(self):
        self.variables = {}

    def publish(self, name, var):
        if name in self.variables:
            raise RuntimeError("'%s' variable have been already published before" % name)
        self.variables[name] = var
        return var

    def get(self, name):
        if name not in self.variables:
            raise RuntimeError("No such variable: %s", name)
        return self.variables[name]

    def get_dict(self):
        return {k: v.get() for k, v in self.variables.items()}


class Var(object):
    """
    This class stores generic variable value.
    It is also a base class for other variable types
    """

    def __init__(self, value=None):
        self.value = value

    def set(self, value):
        self.value = value

    def get(self):
        return self.value

    def __str__(self):
        return str(self.value)


class Int(Var):
    def __init__(self, value=0):
        if not isinstance(value, int):
            raise ValueError("Value should be an integer, but it is '%s'" % type(value))
        super(Int, self).__init__(value)

    def inc(self, delta=1):
        self.value += delta


class Metric(object):
    """
    This class stores generic time-series data in a queue.
    Values are stored as (timestamp, value) tuples
    """

    def __init__(self):
        self.metric = Queue()

    def push(self, value, timestamp=None):
        if timestamp is None:
            timestamp = int(time.time())
        elif not isinstance(timestamp, int):
            raise ValueError("Timestamp should be an integer, but it is '%s'" % type(timestamp))
        self.metric.put((timestamp, value))

    def __next__(self):
        try:
            return self.metric.get_nowait()
        except Empty:
            raise StopIteration

    def get(self):
        # TODO: decide what we should return here
        return None

    def __iter__(self):
        return self


EV = ExpVar()


def publish(name, var):
    return EV.publish(name, var)


def get(name):
    return EV.get(name)


def get_dict():
    return EV.get_dict()



# coding=utf-8
import threading

import pandas as pd

try:
    import queue
except ImportError:
    import Queue as queue
import uuid
import numpy as np
import logging

logger = logging.getLogger(__name__)


class Aggregated(object):
    buffer_size = 10  # seconds

    @classmethod
    def is_aggregated(cls):
        return True


class DataType(object):
    table_name = ''
    columns = []

    @classmethod
    def processor(cls, df):
        """
        :type df: pandas.DataFrame
        :rtype: pandas.DataFrame
        """
        return df

    @classmethod
    def is_aggregated(cls):
        return False


class TypeTimeSeries(DataType):
    table_name = 'metrics'
    columns = ['ts', 'value']


class TypeEvents(DataType):
    table_name = 'events'
    columns = ['ts', 'value']


class TypeQuantiles(Aggregated, DataType):
    perc_list = [0, 10, 25, 50, 75, 80, 85, 90, 95, 98, 99, 100]
    qlist = ['q%d' % n for n in perc_list]
    rename = {'count': 'cnt', 'mean': 'average', 'std': 'stddev', '0%': 'q0', '10%': 'q10', '25%': 'q25',
              '50%': 'q50', '75%': 'q75', '80%': 'q80', '85%': 'q85', '90%': 'q90',
              '95%': 'q95', '98%': 'q98', '99%': 'q99', '100%': 'q100', }

    table_name = 'aggregates'
    columns = ['ts'] + qlist + ['average', 'stddev', 'sum', 'cnt']
    __aggregator_buffer = {}
    aggregator_buffer_size = 10

    @classmethod
    def processor(cls, df, groupby='second'):
        # result = pd.DataFrame.from_dict({ts: self.aggregates(df) for ts, df in by_second.items()}
        #                                 , orient='index', columns=Aggregate.columns)
        if df.empty:
            logging.debug('Empty df for quantiles, skip %s', df)
            return pd.DataFrame()
        df = df.set_index(groupby)
        series = df.loc[:, AbstractMetric.VALUE_COL]
        res = series.groupby(series.index). \
            describe(percentiles=[i / 100. for i in cls.perc_list]). \
            rename(columns=cls.rename)
        res['ts'] = res.index
        res['cnt'] = res['cnt'].astype(int)
        res['sum'] = res['average'] * res['cnt']
        return res


class TypeDistribution(Aggregated, DataType):
    table_name = 'distributions'
    columns = ['ts', 'l', 'r', 'cnt']
    DEFAULT_BINS = np.concatenate((
        np.linspace(0, 4990, 500, dtype=int),  # 10µs accuracy
        np.linspace(5000, 9900, 50, dtype=int),  # 100µs accuracy
        np.linspace(10, 499, 490, dtype=int) * 1000,  # 1ms accuracy
        np.linspace(500, 2995, 500, dtype=int) * 1000,  # 5ms accuracy
        np.linspace(3000, 9990, 700, dtype=int) * 1000,  # 10ms accuracy
        np.linspace(10000, 29950, 400, dtype=int) * 1000,  # 50ms accuracy
        np.linspace(30000, 119900, 900, dtype=int) * 1000,  # 100ms accuracy
        np.linspace(120, 300, 181, dtype=int) * 1000000  # 1s accuracy
    ))

    @classmethod
    def processor(cls, df, bins=DEFAULT_BINS, groupby='second'):
        if df.empty:
            logger.debug('Empty df for distribution, skip %s', df)
            return pd.DataFrame()
        df = df.set_index(groupby)
        series = df.loc[:, AbstractMetric.VALUE_COL]
        data = {ts: np.histogram(s, bins=bins) for ts, s in series.groupby(series.index)}
        # data, bins = np.histogram(series, bins=bins)
        result = pd.concat(
            [pd.DataFrame({'l': bins[:-1],
                           'r': bins[1:],
                           'cnt': cnt,
                           'ts': ts},
                          columns=['l', 'r', 'cnt', 'ts']
                          ).query('cnt > 0') for ts, (cnt, bins) in data.items()]
        )
        return result


class TypeHistogram(Aggregated, DataType):
    table_name = 'histograms'
    columns = ['ts', 'category', 'cnt']

    @classmethod
    def processor(cls, df, groupby='second'):
        if df.empty:
            logger.debug('Empty df for histogram, skip %s', df)
            return pd.DataFrame()
        df = df.set_index(groupby)
        series = df.loc[:, AbstractMetric.VALUE_COL]
        data = series.groupby([series.index, series.values]).size().reset_index(). \
            rename(columns={'second': 'ts', 'level_1': 'category', 'value': 'cnt'})
        return data


class AbstractClient(object):
    def __init__(self, meta, data_session):
        self.local_id = "client_{uuid}".format(uuid=uuid.uuid4())
        self.pending_metrics = []
        self.data_session = data_session
        self.pending_queue = queue.Queue()
        self.meta = meta

    def subscribe(self, metric):
        self.pending_metrics.append(metric)

    def put(self, data_type, df):
        self.pending_queue.put((data_type, df))

    def update_job(self, meta):
        pass

    def update_metric(self, meta):
        pass


class MetricData(object):
    def __init__(self, df, data_types, local_id, test_start):
        """

        :param df: pandas.DataFrame
        :param data_types: list of DataType
        :param local_id: uuid4
        """
        df.loc[:, 'metric_local_id'] = local_id
        df = df.set_index('metric_local_id')
        self.data_types = data_types
        self.local_id = local_id
        df.loc[:, 'ts'] = (df['ts'] - test_start).astype(int)
        if self.is_aggregated:
            df.loc[:, 'second'] = (df['ts'] / 1000000).astype(int)
        self.df = df

    @property
    def is_aggregated(self):
        return any([dtype.is_aggregated() for dtype in self.data_types])

    def __repr__(self):
        return "MetricData: aggregated={}, data types={}\n{}".format(
            'yes' if self.is_aggregated else 'no',
            self.data_types.__repr__(),
            self.df.__repr__()
        )


class AbstractMetric(object):
    VALUE_COL = 'value'
    TS_COL = 'ts'

    def __init__(self, meta, _queue, test_start, raw=True, aggregate=False, parent=None, case=None):
        self.local_id = str(uuid.uuid4())
        self.meta = meta
        self.routing_queue = _queue
        self.test_start = test_start
        self.raw = raw
        self.aggregate = aggregate
        self.parent = parent
        self.case = case
        if not (raw or aggregate):
            raise ValueError('Either raw or aggregate must be True to upload some data')

    @property
    def type(self):
        """
        :rtype: DataType
        """
        raise NotImplementedError('Abstract type property should be redefined!')

    @property
    def aggregate_types(self):
        """
        :rtype: list of DataType
        """
        raise NotImplementedError('Abstract type property should be redefined!')

    @property
    def data_types(self):
        """
        :rtype: list of DataType
        """
        return [self.type] * self.raw + self.aggregate_types * self.aggregate

    def put(self, df):
        # FIXME check dtypes of an incoming dataframe
        data = MetricData(df, self.data_types, self.local_id, self.test_start)
        self.routing_queue.put(data)


class QueueWorker(threading.Thread):
    """ Process data """

    def __init__(self, _queue):
        """
        :type _queue: queue.Queue
        """
        super(QueueWorker, self).__init__()
        self.queue = _queue
        self._finished = threading.Event()
        self._stopped = threading.Event()
        self._interrupted = threading.Event()

    def stop(self):
        self._stopped.set()

    def interrupt(self):
        self._stopped.set()
        self._interrupted.set()

    def run(self):
        while not self._stopped.is_set():
            self._process_pending_queue()
        while self.queue.qsize() > 0 and not self._interrupted.is_set():
            self._process_pending_queue(progress=True)
        while self.queue.qsize() > 0:
            self.queue.get_nowait()
        self._finished.set()

    def _process_pending_queue(self, progress=False):
        raise NotImplementedError

    def is_finished(self):
        return self._finished.is_set()



""" Contains Phantom Plugin, Console widgets, result reader classes """
# FIXME: 3 there is no graceful way to interrupt the process of phout import
# TODO: phout import
import logging
import subprocess
import time
from threading import Event

from .reader import PhantomReader, PhantomStatsReader, string_to_df
from .utils import PhantomConfig
from .widget import PhantomInfoWidget, PhantomProgressBarWidget
from ..Console import Plugin as ConsolePlugin
from ...common.interfaces import GeneratorPlugin
from ...common.util import FileMultiReader
from .log_analyzer import LogAnalyzer

from yandextank.contrib.netort.netort.process import execute

logger = logging.getLogger(__name__)


class Plugin(GeneratorPlugin):
    """     Plugin for running phantom tool    """

    OPTION_CONFIG = "config"
    SECTION = "phantom"

    def __init__(self, core, cfg, name):
        super(Plugin, self).__init__(core, cfg, name)
        self.phout_finished = Event()
        self.predefined_phout = None
        self.did_phout_import_try = False
        self.eta_file = None
        self.processed_ammo_count = 0
        self.cached_info = None
        self.exclude_markers = []
        self._stat_log = None
        self._phantom = None
        self.config = None
        self.enum_ammo = None
        self.phout_import_mode = None
        self.start_time = None

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        opts = [
            "phantom_path", "buffered_seconds", "exclude_markers", "affinity"
        ]
        opts += [PhantomConfig.OPTION_PHOUT, self.OPTION_CONFIG]
        opts += PhantomConfig.get_available_options()
        return opts

    def configure(self):
        # plugin part
        self.config = self.get_option(self.OPTION_CONFIG, '')
        self.affinity = self.get_option('affinity', '')
        self.enum_ammo = self.get_option("enum_ammo", False)
        self.buffered_seconds = int(
            self.get_option("buffered_seconds", self.buffered_seconds))

        self.predefined_phout = self.get_option(PhantomConfig.OPTION_PHOUT, '')
        if not self.get_option(self.OPTION_CONFIG, '') and self.predefined_phout:
            self.phout_import_mode = True
        self.phantom_config = self.phantom.config_file

    @property
    def phantom(self):
        """
        :rtype: PhantomConfig
        """
        if not self._phantom:
            self._phantom = PhantomConfig(self.core, self.cfg, self.stat_log)
            self._phantom.read_config()
        return self._phantom

    @property
    def stat_log(self):
        if not self._stat_log:
            self._stat_log = self.core.mkstemp(".log", "phantom_stat_")
        return self._stat_log

    def get_reader(self, parser=string_to_df):
        if self.reader is None:
            self.reader = FileMultiReader(self.phantom.phout_file, self.phout_finished)
        return PhantomReader(self.reader.get_file(), parser=parser)

    def get_stats_reader(self):
        if self.stats_reader is None:
            self.stats_reader = PhantomStatsReader(self.stat_log, self.phantom.get_info(), lambda: self.start_time)
        return self.stats_reader

    def prepare_test(self):
        try:
            retcode, stdout, stderr = execute(
                [self.get_option("phantom_path"), 'check', self.phantom.config_file],
                catch_out=True,
                env=dict(OPENSSL_CONF="/dev/null")
            )
        except OSError:
            logger.debug("Phantom I/O engine is not installed!", exc_info=True)
            raise OSError("Phantom I/O engine not found. \nMore information: {doc_url}".format(
                doc_url='http://yandextank.readthedocs.io/en/latest/install.html')
            )
        else:
            if retcode or stderr:
                raise RuntimeError("Config check failed. Subprocess returned code %s. Stderr: %s" % (retcode, stderr))

        logger.debug(
            "Linking sample reader to aggregator."
            " Reading samples from %s", self.phantom.phout_file)

        logger.debug(
            "Linking stats reader to aggregator."
            " Reading stats from %s", self.phantom.stat_log)

        self.core.job.aggregator.add_result_listener(self)

        # stepping inside get_info()
        self.core.job.phantom_info = self.phantom.get_info()

        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except KeyError as ex:
            logger.debug(ex)
        else:
            widget1 = PhantomProgressBarWidget(self)
            console.add_info_widget(widget1)
            self.core.job.aggregator.add_result_listener(widget1)

            widget2 = PhantomInfoWidget(self)
            console.add_info_widget(widget2)
            self.core.job.aggregator.add_result_listener(widget2)

    def start_test(self):
        args = [self.get_option("phantom_path"), 'run', self.phantom.config_file]
        if self.affinity:
            logger.info('Enabled cpu affinity %s for phantom', self.affinity)
            args = self.core.__setup_affinity(self.affinity, args=args)
        logger.debug("Starting %s with arguments: %s", self.get_option("phantom_path"), args)
        phantom_stderr_file = self.core.mkstemp(".log", "phantom_stdout_stderr_")
        self.core.add_artifact_file(phantom_stderr_file)
        self.process_stderr = open(phantom_stderr_file, 'w')
        self.start_time = time.time()
        self.process = subprocess.Popen(
            args,
            stderr=self.process_stderr,
            stdout=self.process_stderr,
            close_fds=True,
            env=dict(OPENSSL_CONF="/dev/null"))

    def is_test_finished(self):
        retcode = self.process.poll()
        if retcode is not None:
            logger.info("Phantom done its work with exit code: %s", retcode)
            self.phout_finished.set()
            if retcode != 0:
                errors = LogAnalyzer(self.phantom.phantom_log).get_most_recent_errors()
                if not errors:
                    logger.error('Phantom exited with code %s but without errors in log.')
                self.errors.extend(errors)
            return abs(retcode)
        else:
            info = self.get_info()
            if info:
                eta = int(info.duration) - (int(time.time()) - int(self.start_time))
                self.publish('eta', eta)
            return -1

    def end_test(self, retcode):
        if self.process and self.process.poll() is None:
            logger.info("Terminating phantom process with PID %s", self.process.pid)
            self.process.terminate()
            if self.process:
                self.process.communicate()
        else:
            logger.info("Seems phantom finished")
        self.phout_finished.set()
        if self.process_stderr:
            self.process_stderr.close()
        return retcode

    def post_process(self, retcode):
        if not retcode:
            info = self.get_info()
            if info and info.ammo_count != self.processed_ammo_count:
                logger.warning(
                    "Planned ammo count %s differs from processed %s",
                    info.ammo_count, self.processed_ammo_count)
        return retcode

    def on_aggregated_data(self, data, stat):
        self.processed_ammo_count += data["overall"]["interval_real"]["len"]
        logger.debug("Processed ammo count: %s/", self.processed_ammo_count)

    def get_info(self):
        """ returns info object """
        if not self.cached_info:
            if not self.phantom:
                return None
            self.cached_info = self.phantom.get_info()
        return self.cached_info



import collections.abc
import functools
import inspect
import os
import socket
import shutil

import subprocess
import time
import traceback
from typing import List, Optional

import http.client
import logging
import errno
import re

import psutil
try:
    import pathlib
except ImportError:
    import pathlib2 as pathlib

from retrying import retry

try:
    from library.python import resource as rs
    pip = False
except ImportError:
    pip = True

logger = logging.getLogger(__name__)


def read_resource(path, file_open_mode='r'):
    if not pip and path in rs.iterkeys(prefix='resfs/file/load/projects/yandex-tank/'):
        return rs.find(path).decode('utf8')
    else:
        with open(path, file_open_mode) as f:
            return f.read()


class SecuredShell(object):
    valid_key: str | None = None

    def __init__(self, host, port, username, command_timeout=30, file_timeout=300, protocol_timeout=15, ssh_key_path=None):
        self.connection_address = f'{username}@{host}'
        self.port = port
        self._command_timeout = command_timeout
        self._file_timeout = file_timeout
        self._protocol_timeout = protocol_timeout
        self.key_filename = get_ssh_key_filenames(ssh_key_path=ssh_key_path) or None
        self.default_key_filename = get_ssh_key_filenames(ssh_key_path='~/.ssh/') or []

    def _pick_ssh_key(self):
        key_filename = self.default_key_filename
        if self.key_filename is not None:
            key_filename = self.key_filename
        for filename in key_filename:
            _, _, exit_code = self.execute(
                cmd='exit',
                ssh_opts=[
                    '-i',
                    filename,
                    *self._make_ssh_opts(),
                ])
            if not exit_code:
                return filename
        logger.info('Could not find appropriate file with ssh key')
        return None

    @staticmethod
    def popen(cmd):
        env = os.environ.copy()
        return subprocess.Popen(
            cmd,
            shell=True,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=env,
        )

    @staticmethod
    def check_executable_present(util: str):
        def wrapper(func):
            def inner_wrapper(*args, **kwargs):
                executable_exists = shutil.which(util)
                if executable_exists is None:
                    raise FileNotFoundError(f'{util} executable should be installed to call {func.__name__} on SecureShell')
                return func(*args, **kwargs)
            return inner_wrapper
        return wrapper

    def _make_ssh_opts(self, util: str = 'ssh'):
        port_flag = '-p' if util == 'ssh' else '-P'
        ssh_opts = [
            port_flag,
            str(self.port),
            '-o',
            f'ConnectTimeout={self._protocol_timeout}',
            '-o',
            f'ServerAliveInterval={(self._protocol_timeout + 2) // 3}',
            '-o',
            'ServerAliveCountMax=3',
            '-o',
            'StrictHostKeyChecking=no',
            '-o',
            'BatchMode=yes',
        ]
        if self.valid_key is not None:
            ssh_opts = ['-i', self.valid_key] + ssh_opts
        return ssh_opts

    def _safe_communicate(
        self,
        process: subprocess.Popen,
        timeout: float,
        error_message: str,
    ) -> tuple[bytes, bytes, int]:
        outs, errs, exit_code = b'', b'', 0
        try:
            outs, errs = process.communicate(timeout=timeout)
            exit_code = process.returncode
        except subprocess.TimeoutExpired:
            process.kill()
            logger.error(error_message)
            exit_code = 1
        return outs, errs, exit_code

    def ensure_connection(self):
        self.valid_key = self._pick_ssh_key()
        _, stderr, exit_code = self.execute(cmd='exit')
        if exit_code:
            stderr = stderr or 'Unhandled error.'
            raise ConnectionError(f'Some error occurred in attempt to establish SSH connection: {stderr}')

    @check_executable_present('scp')
    def send_file(self, local_path: str, remote_path: str):
        ssh_opts = self._make_ssh_opts(util='scp')
        full_remote_path = f'{self.connection_address}:{remote_path}'

        cmd = ['scp'] + ssh_opts + [local_path, full_remote_path]
        logger.info('Sending from [%s] to %s:[%s]', local_path, self.connection_address, remote_path)
        process = self.popen(' '.join(cmd))
        _, stderr, exit_code = self._safe_communicate(
            process,
            self._file_timeout,
            f'Timeout while trying to send file from {local_path} to {self.connection_address}:{remote_path}',
        )
        if exit_code:
            raise ConnectionError(f'Some error occurred in attempt to send file: {stderr}')

    @check_executable_present('scp')
    def get_file(self, remote_path: str, local_path: str):
        ssh_opts = self._make_ssh_opts(util='scp')
        full_remote_path = f'{self.connection_address}:{remote_path}'

        cmd = ['scp'] + ssh_opts + [full_remote_path, local_path]
        logger.info('Receiving from %s:[%s] to [%s]', self.connection_address, remote_path, local_path)
        process = self.popen(' '.join(cmd))
        _, stderr, exit_code = self._safe_communicate(
            process,
            self._file_timeout,
            f'Timeout while trying to get file from {self.connection_address}:{remote_path} to {local_path}',
        )
        if exit_code:
            raise ConnectionError(f'Some error occurred in attempt to get file: {stderr}')

    @check_executable_present('ssh')
    def execute(self, cmd: str, ssh_opts: Optional[List[str]] = None):
        ssh_opts = ssh_opts or self._make_ssh_opts()
        ssh_cmd = ['ssh'] + ssh_opts + [self.connection_address, cmd]

        logger.info('Executing: %s', ssh_cmd)
        process = self.popen(' '.join(ssh_cmd))
        stdout, stderr, exit_code = self._safe_communicate(
            process, self._command_timeout, f'Timeout while trying to execute command {ssh_cmd}'
        )

        stdout = stdout.decode('utf-8')
        stderr = stderr.decode('utf-8')

        if exit_code:
            logger.error(f'Failed to execute command with exit code {exit_code}\nSTDOUT:\n{stdout}\nSTDERR:\n{stderr}')
        return stdout, stderr, exit_code

    @check_executable_present('ssh')
    def execute_without_communicate(self, cmd: str):
        ssh_opts = self._make_ssh_opts()
        ssh_cmd = ['ssh'] + ssh_opts + [self.connection_address, cmd]
        return self.popen(' '.join(ssh_cmd))

    def rm_r(self, path: str):
        if not path or path == '/':
            raise ValueError('Directory is empty or root directory')
        return self.execute(f'rm -rf {path}')

    def async_session(self, cmd: str):
        return Session(self, cmd)


class Session:
    def __init__(self, client: SecuredShell, cmd: str):
        self.client = client
        self.process = self.client.execute_without_communicate(cmd)
        self._exit_status: int | None = None
        self.stdout = self.process.stdout
        os.set_blocking(self.stdout.fileno(), False)

    def send(self, data):
        if not self.is_finished() and self.process.stdin:
            self.process.stdin.write(data)
            self.process.stdin.flush()

    def read_maybe(self):
        output = self.stdout.read(4096)
        if output:
            return output.decode('utf-8')
        return None

    def is_finished(self):
        return self.exit_status() is not None

    def exit_status(self):
        if self._exit_status is None:
            self._exit_status = self.process.poll()
        return self._exit_status

    def close(self, timeout=10):
        try:
            self.process.communicate(timeout=timeout)
        except subprocess.TimeoutExpired:
            logger.warning('Process has not been ended until timeout expired')
            self.process.kill()
        self.exit_status()


# HTTP codes
HTTP = http.client.responses

# Extended list of HTTP status codes(WEBdav etc.)
# HTTP://en.wikipedia.org/wiki/List_of_HTTP_status_codes
WEBDAV = {
    102: 'Processing',
    103: 'Checkpoint',
    122: 'Request-URI too long',
    207: 'Multi-Status',
    226: 'IM Used',
    308: 'Resume Incomplete',
    418: 'I\'m a teapot',
    422: 'Unprocessable Entity',
    423: 'Locked',
    424: 'Failed Dependency',
    425: 'Unordered Collection',
    426: 'Upgrade Required',
    444: 'No Response',
    449: 'Retry With',
    450: 'Blocked by Windows Parental Controls',
    499: 'Client Closed Request',
    506: 'Variant Also Negotiates',
    507: 'Insufficient Storage',
    509: 'Bandwidth Limit Exceeded',
    510: 'Not Extended',
    598: 'network read timeout error',
    599: 'network connect timeout error',
    999: 'Common Failure',
}
HTTP.update(WEBDAV)

# NET codes
NET = {
    0: "Success",
    1: "Operation not permitted",
    2: "No such file or directory",
    3: "No such process",
    4: "Interrupted system call",
    5: "Input/output error",
    6: "No such device or address",
    7: "Argument list too long",
    8: "Exec format error",
    9: "Bad file descriptor",
    10: "No child processes",
    11: "Resource temporarily unavailable",
    12: "Cannot allocate memory",
    13: "Permission denied",
    14: "Bad address",
    15: "Block device required",
    16: "Device or resource busy",
    17: "File exists",
    18: "Invalid cross-device link",
    19: "No such device",
    20: "Not a directory",
    21: "Is a directory",
    22: "Invalid argument",
    23: "Too many open files in system",
    24: "Too many open files",
    25: "Inappropriate ioctl for device",
    26: "Text file busy",
    27: "File too large",
    28: "No space left on device",
    29: "Illegal seek",
    30: "Read-only file system",
    31: "Too many links",
    32: "Broken pipe",
    33: "Numerical argument out of domain",
    34: "Numerical result out of range",
    35: "Resource deadlock avoided",
    36: "File name too long",
    37: "No locks available",
    38: "Function not implemented",
    39: "Directory not empty",
    40: "Too many levels of symbolic links",
    41: "Unknown error 41",
    42: "No message of desired type",
    43: "Identifier removed",
    44: "Channel number out of range",
    45: "Level 2 not synchronized",
    46: "Level 3 halted",
    47: "Level 3 reset",
    48: "Link number out of range",
    49: "Protocol driver not attached",
    50: "No CSI structure available",
    51: "Level 2 halted",
    52: "Invalid exchange",
    53: "Invalid request descriptor",
    54: "Exchange full",
    55: "No anode",
    56: "Invalid request code",
    57: "Invalid slot",
    58: "Unknown error 58",
    59: "Bad font file format",
    60: "Device not a stream",
    61: "No data available",
    62: "Timer expired",
    63: "Out of streams resources",
    64: "Machine is not on the network",
    65: "Package not installed",
    66: "Object is remote",
    67: "Link has been severed",
    68: "Advertise error",
    69: "Srmount error",
    70: "Communication error on send",
    71: "Protocol error",
    72: "Multihop attempted",
    73: "RFS specific error",
    74: "Bad message",
    75: "Value too large for defined data type",
    76: "Name not unique on network",
    77: "File descriptor in bad state",
    78: "Remote address changed",
    79: "Can not access a needed shared library",
    80: "Accessing a corrupted shared library",
    81: ".lib section in a.out corrupted",
    82: "Attempting to link in too many shared libraries",
    83: "Cannot exec a shared library directly",
    84: "Invalid or incomplete multibyte or wide character",
    85: "Interrupted system call should be restarted",
    86: "Streams pipe error",
    87: "Too many users",
    88: "Socket operation on non-socket",
    89: "Destination address required",
    90: "Message too long",
    91: "Protocol wrong type for socket",
    92: "Protocol not available",
    93: "Protocol not supported",
    94: "Socket type not supported",
    95: "Operation not supported",
    96: "Protocol family not supported",
    97: "Address family not supported by protocol",
    98: "Address already in use",
    99: "Cannot assign requested address",
    100: "Network is down",
    101: "Network is unreachable",
    102: "Network dropped connection on reset",
    103: "Software caused connection abort",
    104: "Connection reset by peer",
    105: "No buffer space available",
    106: "Transport endpoint is already connected",
    107: "Transport endpoint is not connected",
    108: "Cannot send after transport endpoint shutdown",
    109: "Too many references: cannot splice",
    110: "Connection timed out",
    111: "Connection refused",
    112: "Host is down",
    113: "No route to host",
    114: "Operation already in progress",
    115: "Operation now in progress",
    116: "Stale NFS file handle",
    117: "Structure needs cleaning",
    118: "Not a XENIX named type file",
    119: "No XENIX semaphores available",
    120: "Is a named type file",
    121: "Remote I/O error",
    122: "Disk quota exceeded",
    123: "No medium found",
    124: "Wrong medium type",
    125: "Operation canceled",
    126: "Required key not available",
    127: "Key has expired",
    128: "Key has been revoked",
    129: "Key was rejected by service",
    130: "Owner died",
    131: "State not recoverable",
    999: 'Common Failure',
}


def expand_to_milliseconds(str_time):
    """
    converts 1d2s into milliseconds
    """
    return expand_time(str_time, 'ms', 1000)


def expand_to_seconds(str_time):
    """
    converts 1d2s into seconds
    """
    return expand_time(str_time, 's', 1)


def expand_time(str_time, default_unit='s', multiplier=1):
    """
    helper for above functions
    """
    parser = re.compile(r'(\d+)([a-zA-Z]*)')
    parts = parser.findall(str_time)
    result = 0.0
    for value, unit in parts:
        value = int(value)
        unit = unit.lower()
        if unit == '':
            unit = default_unit

        if unit == 'ms':
            result += value * 0.001
            continue
        elif unit == 's':
            result += value
            continue
        elif unit == 'm':
            result += value * 60
            continue
        elif unit == 'h':
            result += value * 60 * 60
            continue
        elif unit == 'd':
            result += value * 60 * 60 * 24
            continue
        elif unit == 'w':
            result += value * 60 * 60 * 24 * 7
            continue
        else:
            raise ValueError("String contains unsupported unit %s: %s" % (unit, str_time))
    return int(result * multiplier)


def pid_exists(pid):
    """Check whether pid exists in the current process table."""
    if pid < 0:
        return False
    try:
        os.kill(pid, 0)
    except OSError as exc:
        logging.debug("No process[%s]: %s", exc.errno, exc)
        return exc.errno == errno.EPERM
    else:
        p = psutil.Process(pid)
        return p.status != psutil.STATUS_ZOMBIE


def pairs(lst):
    """
    Iterate over pairs in the list
    """
    return zip(lst[::2], lst[1::2])


class AddressWizard:
    def __init__(self):
        self.lookup_fn = socket.getaddrinfo
        self.socket_class = socket.socket

    def resolve(self, address_str, do_test=False, explicit_port=False):
        """

        :param address_str:
        :return: tuple of boolean, string, int - isIPv6, resolved_ip, port (may be null), extracted_address
        """

        if not address_str:
            raise RuntimeError("Mandatory option was not specified: address")

        logger.debug("Trying to resolve address string: %s", address_str)

        port = None

        braceport_re = re.compile(r"""
            ^
            \[           # opening brace
            \s?          # space sym?
            (\S+)        # address - string
            \s?          # space sym?
            \]           # closing brace
            :            # port separator
            \s?          # space sym?
            (\d+)        # port
            $
        """, re.X)
        braceonly_re = re.compile(r"""
            ^
            \[           # opening brace
            \s?          # space sym?
            (\S+)        # address - string
            \s?          # space sym?
            \]           # closing brace
            $
        """, re.X)

        if braceport_re.match(address_str):
            logger.debug("Braces and port present")
            match = braceport_re.match(address_str)
            logger.debug("Match: %s %s ", match.group(1), match.group(2))
            address_str, port = match.group(1), match.group(2)
        elif braceonly_re.match(address_str):
            logger.debug("Braces only present")
            match = braceonly_re.match(address_str)
            logger.debug("Match: %s", match.group(1))
            address_str = match.group(1)
        else:
            logger.debug("Parsing port")
            parts = address_str.split(":")
            if len(parts) <= 2:  # otherwise it is v6 address
                address_str = parts[0]
                if len(parts) == 2:
                    port = int(parts[1])
        if port is not None:
            port = int(port)
        address_str = address_str.strip()
        try:
            resolved = self.lookup_fn(address_str, port)
            logger.debug("Lookup result: %s", resolved)
        except Exception:
            logger.debug("Exception trying to resolve hostname %s :", address_str, exc_info=True)
            raise

        for (family, socktype, proto, canonname, sockaddr) in resolved:
            is_v6 = family == socket.AF_INET6
            parsed_ip, port = sockaddr[0], sockaddr[1]

            if explicit_port:
                logger.warning("Using phantom.port option is deprecated. Use phantom.address=[address]:port instead")
                port = int(explicit_port)
            elif not port:
                port = 80

            if do_test:
                try:
                    logger.info("Testing connection to resolved address %s and port %s", parsed_ip, port)
                    self.__test(family, (parsed_ip, port))
                except RuntimeError:
                    logger.info("Failed TCP connection test using [%s]:%s", parsed_ip, port)
                    logger.debug("Failed TCP connection test using [%s]:%s", parsed_ip, port, exc_info=True)
                    continue
            return is_v6, parsed_ip, int(port), address_str

        msg = "All connection attempts failed for %s, use {phantom.connection_test: false} to disable it"
        raise RuntimeError(msg % address_str)

    def __test(self, af, sa):
        test_sock = self.socket_class(af)
        try:
            test_sock.settimeout(5)
            test_sock.connect(sa)
        except Exception:
            logger.debug(
                "Exception on connect attempt [%s]:%s : %s", sa[0], sa[1],
                traceback.format_exc())
            msg = "TCP Connection test failed for [%s]:%s, use phantom.connection_test=0 to disable it"
            raise RuntimeError(msg % (sa[0], sa[1]))
        finally:
            test_sock.close()


def recursive_dict_update(d1, d2):
    # the actual field may be of union type, as in telegraf and pandora plugins: [dict, string]
    if not isinstance(d1, collections.abc.MutableMapping):
        return d2 if d2 is not None else d1
    for k, v in d2.items():
        if isinstance(v, collections.abc.Mapping):
            r = recursive_dict_update(d1.get(k, {}), v)
            d1[k] = r
        else:
            d1[k] = d2[k]
    return d1


class FileScanner(object):
    """
    Basic class for stats reader for continiuos reading file line by line

    Default line separator is a newline symbol. You can specify other separator
    via constructor argument
    """

    _BUFSIZE = 4096

    def __init__(self, path, sep="\n"):
        self.__path = path
        self.__sep = sep
        self.__closed = False
        self.__buffer = ""

    def _read_lines(self, chunk):
        self.__buffer += chunk
        portions = self.__buffer.split(self.__sep)
        for portion in portions[:-1]:
            yield portion
        self.__buffer = portions[-1]

    def _read_data(self, lines):
        raise NotImplementedError()

    def __iter__(self):
        with open(self.__path) as stats_file:
            while not self.__closed:
                chunk = stats_file.read(self._BUFSIZE)
                yield self._read_data(self._read_lines(chunk))

    def close(self):
        self.__closed = True


def tail_lines(filepath, lines_num, bufsize=8192):
    fsize = os.stat(filepath).st_size
    logging.warning('Filepath=%s, lines_num=%s, buf_size=%s, fsize=%s', filepath, lines_num, bufsize, fsize)

    iter_ = 0
    with open(filepath) as f:
        if bufsize > fsize:
            bufsize = fsize - 1
        data = []
        try:
            while True:
                iter_ += 1
                line_start_pos = max(0, fsize - bufsize * iter_)
                f.seek(line_start_pos)
                data.extend(f.readlines())
                if len(data) >= lines_num or f.tell() == 0:
                    return data[-lines_num:]
        except (IOError, OSError):
            return data


class FileLockedError(RuntimeError):
    pass

    @classmethod
    def retry(cls, exception):
        return isinstance(exception, cls)


class FileMultiReader(object):
    def __init__(self, filename, provider_stop_event, cache_size=1024 * 1024 * 50):
        self.buffer = ""
        self.filename = filename
        self.cache_size = cache_size
        self._cursor_map = {}
        self._is_locked = False
        self._opened_file = open(self.filename)
        self.stop = provider_stop_event

    def close(self, force=False):
        self.wait_lock()
        self._opened_file.close()
        self.unlock()

    def get_file(self, cache_size=None):
        cache_size = self.cache_size if not cache_size else cache_size
        fileobj = FileLike(self, cache_size)
        return fileobj

    def read_with_lock(self, pos, _len=None):
        """
        Reads {_len} characters if _len is not None else reads line
        :param pos: start reading position
        :param _len: number of characters to read
        :rtype: (string, int)
        """
        self.wait_lock()
        try:
            self._opened_file.seek(pos)
            result = self._opened_file.read(_len) if _len is not None else self._opened_file.readline()
            stop_pos = self._opened_file.tell()
        finally:
            self.unlock()
        if not result and self.stop.is_set():
            result = None
        return result, stop_pos

    @retry(wait_random_min=5, wait_random_max=20, stop_max_delay=10000,
           retry_on_exception=FileLockedError.retry, wrap_exception=True)
    def wait_lock(self):
        if self._is_locked:
            raise FileLockedError('Generator output file {} is locked'.format(self.filename))
        else:
            self._is_locked = True
            return True

    def unlock(self):
        self._is_locked = False


class FileLike(object):
    def __init__(self, multireader, cache_size):
        """
        :type multireader: FileMultiReader
        """
        self.multireader = multireader
        self.cache_size = cache_size
        self._cursor = 0

    def read(self, _len=None):
        _len = self.cache_size if not _len else _len
        result, self._cursor = self.multireader.read_with_lock(self._cursor, _len)
        return result

    def readline(self):
        result, self._cursor = self.multireader.read_with_lock(self._cursor)
        return result


def get_callstack():
    """
        Get call stack, clean wrapper functions from it and present
        in dotted notation form
    """
    stack = inspect.stack(context=0)
    cleaned = [frame[3] for frame in stack if frame[3] != 'wrapper']
    return '.'.join(cleaned[1:])


def observetime(name=None, log=None):
    log = log or logger
    name_ = name

    def observetime_fixed(func):
        name = name_ or func.__name__

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            start_time = time.time()
            result = func(*args, **kwargs)
            duration = time.time() - start_time
            log.debug('%s completed in %s seconds', name, duration)
            return result
        return wrapper

    return observetime_fixed


def timeit(min_duration_sec):
    def timeit_fixed(func):
        def wrapper(*args, **kwargs):
            start_time = time.time()
            result = func(*args, **kwargs)
            stack = get_callstack()
            duration = time.time() - start_time
            if duration > min_duration_sec:
                logger.warning('Slow call of %s (stack: %s), duration %s', func.__name__, stack, duration)
            return result
        return wrapper
    return timeit_fixed


def for_all_methods(decorator, exclude=None):
    if exclude is None:
        exclude = []
    """
        Decorator for all methods in a class,
        shamelessly stolen from https://stackoverflow.com/questions/6307761
    """
    def decorate(cls):
        for attr in cls.__dict__:  # there's propably a better way to do this
            if callable(getattr(cls, attr)) and attr not in exclude:
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls
    return decorate


class Cleanup:
    def __init__(self, tankworker):
        """

        :type tankworker: TankWorker
        """
        self._actions = []
        self.tankworker = tankworker

    def add_action(self, name, fn):
        """

        :type fn: function
        :type name: str
        """
        assert callable(fn)
        self._actions.append((name, fn))

    def __enter__(self):
        return self.add_action

    def __exit__(self, exc_type, exc_val, exc_tb):
        msgs = []
        if exc_type:
            msg = 'Exception occurred:\n{}: {}\n{}'.format(exc_type, exc_val, '\n'.join(traceback.format_tb(exc_tb)))
            self.tankworker.retcode = 1
            msgs.append(msg)
            logger.error(msg)
        logger.info('Trying to clean up')
        for name, action in reversed(self._actions):
            try:
                action()
            except Exception:
                msg = 'Exception occurred during cleanup action {}'.format(name)
                msgs.append(msg)
                logger.exception(msg)
        self.tankworker.add_msgs(*msgs)
        self.tankworker.save_finish_status()
        self.tankworker.core._collect_artifacts()
        self.tankworker.status = Status.TEST_FINISHED
        self.tankworker.core.close()
        return False  # re-raise exception


class Finish:
    def __init__(self, tankworker):
        """
        :type tankworker: TankWorker
        """
        self.worker = tankworker

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.worker.status = Status.TEST_FINISHING
        retcode = self.worker.retcode
        if exc_type:
            msg = 'Test interrupted:\n{}: {}\n{}'.format(exc_type, exc_val, '\n'.join(traceback.format_tb(exc_tb)))
            logger.error(msg)
            self.worker.add_msgs(msg)
            retcode = 1
        logger.info('TankWorker RC: %s', retcode)
        retcode = self.worker.core.plugins_end_test(retcode)
        self.worker.retcode = retcode
        return True  # swallow exception & proceed to post-processing


class TankapiLogFilter(logging.Filter):
    def filter(self, record):
        return record.name != 'tankapi'


class Status:
    TEST_POST_PROCESS = b'POST_PROCESS'
    TEST_INITIATED = b'INITIATED'
    TEST_PREPARING = b'PREPARING'
    TEST_NOT_FOUND = b'NOT_FOUND'
    TEST_WAITING_FOR_A_COMMAND_TO_RUN = b'WAITING_FOR_A_COMMAND_TO_RUN'
    TEST_RUNNING = b'RUNNING'
    TEST_FINISHING = b'FINISHING'
    TEST_FINISHED = b'FINISHED'


def get_test_path():
    try:
        from yatest import common
        return common.source_path('load/projects/yandex-tank')
    except ImportError:
        return os.path.dirname(os.path.dirname(os.path.dirname(__file__)))


def get_ssh_key_filenames(ssh_key_path: str) -> list[str]:
    if not ssh_key_path:
        return []
    absolute_key_path = pathlib.Path(os.path.expanduser(ssh_key_path))
    try:
        key_filenames = [str(f) for f in absolute_key_path.iterdir() if f.is_file()]
    except Exception as err:
        logger.warning('Could not access keys with %s path : %s', absolute_key_path, err)
        return []
    return key_filenames



""" jmeter load generator support """
import datetime
import logging
import os
import signal
import subprocess
import time
import socket
import re
import shlex

from pkg_resources import resource_string
from yandextank.contrib.netort.netort.resource import TempDownloaderOpenerProtocol

from .reader import JMeterReader
from ..Console import Plugin as ConsolePlugin
from ..Console import screen as ConsoleScreen
from ...common.interfaces import AggregateResultListener, AbstractInfoWidget, GeneratorPlugin

logger = logging.getLogger(__name__)


class Plugin(GeneratorPlugin):
    """ JMeter tank plugin """
    SECTION = 'jmeter'
    SHUTDOWN_TEST = 'Shutdown'
    STOP_TEST_NOW = 'Stop Test'
    DISCOVER_PORT_PATTERN = r'Waiting for possible .* message on port (?P<port>\d+)'

    def __init__(self, core, cfg, name):
        super(Plugin, self).__init__(core, cfg, name)
        self.args = None
        self.original_jmx = None
        self.jtl_file = None
        self.ext_log = None
        self.ext_levels = ['none', 'errors', 'all']
        self.ext_log_file = None
        self.jmx = None
        self.user_args = None
        self.jmeter_path = None
        self.jmeter_ver = None
        self.jmeter_log = None
        self.start_time = time.time()
        self.jmeter_buffer_size = None
        self.jmeter_udp_port = None
        self.shutdown_timeout = None

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        return [
            "jmx", "args", "jmeter_path", "buffer_size", "buffered_seconds",
            "exclude_markers", "shutdown_timeout"
        ]

    def configure(self):
        opener = self.core.resource_manager.get_opener(self.get_option("jmx"))
        if isinstance(opener, TempDownloaderOpenerProtocol):
            self.original_jmx = opener.download_file(True, try_ungzip=True)
        else:
            self.original_jmx = opener.get_filename
        self.core.add_artifact_file(self.original_jmx, True)
        self.jtl_file = self.core.mkstemp('.jtl', 'jmeter_')
        self.core.add_artifact_file(self.jtl_file)
        self.user_args = self.get_option("args", '')
        self.jmeter_path = self.get_option('jmeter_path')
        self.jmeter_log = self.core.mkstemp('.log', 'jmeter_')
        self.jmeter_ver = self.get_option('jmeter_ver')
        self.ext_log = self.get_option('extended_log', self.get_option('ext_log'))

        if self.ext_log != 'none':
            self.ext_log_file = self.core.mkstemp('.jtl', 'jmeter_ext_')
            self.core.add_artifact_file(self.ext_log_file)

        self.core.add_artifact_file(self.jmeter_log, True)
        self.exclude_markers = set(self.get_option('exclude_markers', []))
        self.jmx = self.__add_jmeter_components(
            self.original_jmx, self.jtl_file, self.get_option('variables'))
        self.core.add_artifact_file(self.jmx)

        jmeter_stderr_file = self.core.mkstemp(".log", "jmeter_stdout_stderr_")
        self.core.add_artifact_file(jmeter_stderr_file)
        self.process_stderr = open(jmeter_stderr_file, 'w')
        self.shutdown_timeout = self.get_option('shutdown_timeout', 3)

        self.affinity = self.get_option('affinity', '')

    def get_reader(self):
        if self.reader is None:
            self.reader = JMeterReader(self.jtl_file, self.core.data_poller)
        return self.reader

    def get_stats_reader(self):
        if self.stats_reader is None:
            self.stats_reader = self.get_reader().stats_reader
        return self.stats_reader

    def prepare_test(self):
        self.args = [
            self.jmeter_path, "-n", "-t", self.jmx, '-j', self.jmeter_log,
            '-Jjmeter.save.saveservice.default_delimiter=\\t',
            '-Jjmeter.save.saveservice.connect_time=true'
        ]
        self.args += shlex.split(self.user_args)

        if self.affinity:
            self.core.__setup_affinity(self.affinity, args=self.args)

        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except KeyError as ex:
            logger.debug(ex)
            console = None

        if console:
            widget = JMeterInfoWidget(self)
            console.add_info_widget(widget)
            self.core.job.aggregator.add_result_listener(widget)

    def start_test(self):
        logger.info("Starting %s with arguments: %s", self.jmeter_path, self.args)
        try:
            self.process = subprocess.Popen(
                self.args,
                executable=self.jmeter_path,
                preexec_fn=os.setsid,
                close_fds=True,
                stdout=self.process_stderr,
                stderr=self.process_stderr)
        except OSError:
            logger.debug(
                "Unable to start JMeter process. Args: %s, Executable: %s",
                self.args,
                self.jmeter_path,
                exc_info=True)
            raise RuntimeError(
                "Unable to access to JMeter executable file or it does not exist: %s"
                % self.jmeter_path)
        self.start_time = time.time()
        self.jmeter_udp_port = self.__discover_jmeter_udp_port()

    def is_test_finished(self):
        retcode = self.process.poll()
        aggregator = self.core.job.aggregator
        if not aggregator.reader.jmeter_finished and retcode is not None:
            logger.info("JMeter process finished with exit code: %s, waiting for aggregator", retcode)
            self.retries = 0
            aggregator.reader.jmeter_finished = True
            return -1
        elif aggregator.reader.jmeter_finished is True:
            if aggregator.reader.agg_finished:
                self.reader.close()
                return retcode
            else:
                logger.info("Waiting for aggregator to finish")
                return -1
        else:
            return -1

    def end_test(self, retcode):
        if self.process:
            gracefully_shutdown = self.__graceful_shutdown()
            if not gracefully_shutdown:
                logger.info('JMeter process will be killed')
                self.__kill_jmeter()
        if self.process_stderr:
            self.process_stderr.close()
        self.core.add_artifact_file(self.jmeter_log)
        self.reader.close()
        return retcode

    def __discover_jmeter_udp_port(self):
        """Searching for line in jmeter.log such as
        Waiting for possible shutdown message on port 4445
        """
        r = re.compile(self.DISCOVER_PORT_PATTERN)
        with open(self.process_stderr.name, 'r') as f:
            cnt = 0
            while self.process.pid and cnt < 10:
                line = f.readline()
                m = r.match(line)
                if m is None:
                    cnt += 1
                    time.sleep(1)
                else:
                    port = int(m.group('port'))
                    return port
            else:
                logger.warning('JMeter UDP port wasn\'t discovered')
                return None

    def __kill_jmeter(self):
        logger.info("Terminating jmeter process group with PID %s", self.process.pid)
        try:
            os.killpg(self.process.pid, signal.SIGTERM)
        except OSError as exc:
            logger.info("Seems JMeter exited itself: %s", exc)
            # Utils.log_stdout_stderr(logger, self.process.stdout, self.process.stderr, "jmeter")

    def __add_jmeter_components(self, jmx, jtl, variables):
        """ Genius idea by Alexey Lavrenyuk """
        logger.debug("Original JMX: %s", os.path.realpath(jmx))
        with open(jmx, 'r') as src_jmx:
            source_lines = src_jmx.readlines()

        try:
            # In new Jmeter version (3.2 as example) WorkBench's plugin checkbox enabled by default
            # It totally crashes Yandex tank injection and raises XML Parse Exception
            closing = source_lines.pop(-1)
            if "WorkBenchGui" in source_lines[-5]:
                logger.info("WorkBench checkbox enabled...bypassing")
                last_string_count = 6
            else:
                last_string_count = 2
            while last_string_count > 0:
                closing = source_lines.pop(-1) + closing
                last_string_count -= 1
            logger.debug("Closing statement: %s", closing)
        except Exception as exc:
            raise RuntimeError("Failed to find the end of JMX XML: %s" % exc)

        udv_tpl = resource_string(__name__, 'config/jmeter_var_template.xml').decode('utf8')
        udv_set = []
        for var_name, var_value in variables.items():
            udv_set.append(udv_tpl % (var_name, var_name, var_value))
        udv = "\n".join(udv_set)

        if self.jmeter_ver >= 2.13:
            save_connect = '<connectTime>true</connectTime>'
        else:
            save_connect = ''

        if self.ext_log in ['errors', 'all']:
            level_map = {'errors': 'true', 'all': 'false'}
            tpl_resource = 'jmeter_writer_ext.xml'
            tpl_args = {
                'jtl': self.jtl_file,
                'udv': udv,
                'ext_log': self.ext_log_file,
                'ext_level': level_map[self.ext_log],
                'save_connect': save_connect
            }
        else:
            tpl_resource = 'jmeter_writer.xml'
            tpl_args = {
                'jtl': self.jtl_file,
                'udv': udv,
                'save_connect': save_connect
            }

        tpl = resource_string(__name__, 'config/' + tpl_resource).decode('utf8')

        try:
            new_jmx = self.core.mkstemp('.jmx', 'modified_', os.path.dirname(os.path.realpath(jmx)))
        except OSError as exc:
            logger.debug("Can't create modified jmx near original: %s", exc)
            new_jmx = self.core.mkstemp('.jmx', 'modified_')
        logger.debug("Modified JMX: %s", new_jmx)
        with open(new_jmx, "w") as fh:
            fh.write(''.join(source_lines))
            fh.write(tpl % tpl_args)
            fh.write(closing)
        return new_jmx

    def __graceful_shutdown(self):
        if self.jmeter_udp_port is None:
            return False
        shutdown_test_started = time.time()
        while time.time() - shutdown_test_started < self.shutdown_timeout:
            self.__send_udp_message(self.SHUTDOWN_TEST)
            if self.process.poll() is not None:
                return True
            else:
                time.sleep(1)
        self.log.info('Graceful shutdown failed after %s' % str(time.time() - shutdown_test_started))

        stop_test_started = time.time()
        while time.time() - stop_test_started < self.shutdown_timeout:
            self.__send_udp_message(self.STOP_TEST_NOW)
            if self.process.poll() is not None:
                return True
            else:
                time.sleep(1)
        self.log.info('Graceful stop failed after {}'.format(time.time() - stop_test_started))
        return False

    def __send_udp_message(self, message):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.sendto(message.encode('utf8'), ('localhost', self.jmeter_udp_port))


class JMeterInfoWidget(AbstractInfoWidget, AggregateResultListener):
    """ Right panel widget with JMeter test info """

    def __init__(self, jmeter):
        AbstractInfoWidget.__init__(self)
        self.krutilka = ConsoleScreen.krutilka()
        self.jmeter = jmeter
        self.active_threads = 0
        self.RPS = 0

    def get_index(self):
        return 0

    def on_aggregated_data(self, data, stats):
        self.active_threads = stats['metrics']['instances']
        self.RPS = data['overall']['interval_real']['len']

    def render(self, screen):
        jmeter = " JMeter Test %s" % next(self.krutilka)
        space = screen.right_panel_width - len(jmeter) - 1
        left_spaces = space // 2
        right_spaces = space // 2

        dur_seconds = int(time.time()) - int(self.jmeter.start_time)
        duration = str(datetime.timedelta(seconds=dur_seconds))

        template = screen.markup.BG_MAGENTA + '~' * left_spaces + jmeter + ' '
        template += '~' * right_spaces + screen.markup.RESET + "\n"
        template += "     Test Plan: %s\n"
        template += "      Duration: %s\n"
        template += "Active Threads: %s\n"
        template += "   Responses/s: %s"
        data = (
            os.path.basename(self.jmeter.original_jmx), duration,
            self.active_threads, self.RPS)

        return template % data



""" Core module to calculate aggregate data """
import json
import logging
import queue as q
from datetime import datetime

from pkg_resources import resource_string
from typing import Collection

from .aggregator import Aggregator, DataPoller
from .chopper import TimeChopper
from yandextank.common.interfaces import AggregateResultListener, StatsReader

from yandextank.contrib.netort.netort.data_processing import Drain, Chopper, get_nowait_from_queue

logger = logging.getLogger(__name__)


class LoggingListener(AggregateResultListener):
    """ Log aggregated results """

    def on_aggregated_data(self, data, stats):
        logger.info("Got aggregated sample:\n%s", json.dumps(data, indent=2))
        logger.info("Stats:\n%s", json.dumps(stats, indent=2))


class TankAggregator(object):
    """
    Plugin that manages aggregation and stats collection
    """

    SECTION = 'aggregator'

    @staticmethod
    def get_key():
        return __file__

    def __init__(self, generator, poller: DataPoller, termination_timeout: float = 60):
        # AbstractPlugin.__init__(self, core, cfg)
        """

        :type generator: GeneratorPlugin
        """
        self.generator = generator
        self.listeners = []  # [LoggingListener()]
        self.results = q.Queue()
        self.stats_results = q.Queue()
        self.data_cache = {}
        self.stat_cache = {}
        self.reader = None
        self.stats_reader = None
        self.drain = None
        self.stats_drain = None
        self.termination_timeout = termination_timeout
        self.poller = poller

    @staticmethod
    def load_config():
        return json.loads(resource_string(__name__, 'config/phout.json').decode('utf8'))

    def start_test(self):
        self.reader = self.generator.get_reader()
        self.stats_reader = self.generator.get_stats_reader()
        aggregator_config = self.load_config()

        if self.reader and self.stats_reader:
            readers = self.reader
            if not isinstance(readers, Collection):
                readers = [readers]

            sources = [self.poller.poll(r) for r in readers]
            pipeline = Aggregator(TimeChopper(sources), aggregator_config)
            self.drain = Drain(pipeline, self.results)
            self.drain.start()

            stats_drain_source = Chopper(self.poller.poll(self.stats_reader))
            self.stats_drain = Drain(stats_drain_source, self.stats_results)
            self.stats_drain.start()
        else:
            logger.warning("Generator not found. Generator must provide a reader and a stats_reader interface")

    def _collect_data(self, end=False):
        """
        Collect data, cache it and send to listeners
        """
        data = get_nowait_from_queue(self.results)
        stats = get_nowait_from_queue(self.stats_results)
        logger.debug("Data timestamps: %s" % [d.get('ts') for d in data])
        logger.debug("Stats timestamps: %s" % [d.get('ts') for d in stats])
        for item in data:
            ts = item['ts']
            if ts in self.stat_cache:
                # send items
                data_item = item
                stat_item = self.stat_cache.pop(ts)
                self.__notify_listeners(data_item, stat_item)
            else:
                self.data_cache[ts] = item
        for item in stats:
            ts = item['ts']
            if ts in self.data_cache:
                # send items
                data_item = self.data_cache.pop(ts)
                stat_item = item
                self.__notify_listeners(data_item, stat_item)
            else:
                self.stat_cache[ts] = item
        if end and len(self.data_cache) > 0:
            logger.info('Timestamps without stats:')
            for ts, data_item in sorted(self.data_cache.items(), key=lambda i: i[0]):
                logger.info(ts)
                self.__notify_listeners(data_item, StatsReader.stats_item(ts, 0, 0))

    def is_aggr_finished(self):
        return self.drain._finished.is_set() and self.stats_drain._finished.is_set()

    def is_test_finished(self):
        self._collect_data()
        return -1

    def end_test(self, retcode):
        timeouter = _Timeouter(self.termination_timeout)
        retcode = self.generator.end_test(retcode)
        if self.stats_reader:
            logger.info('Closing stats reader')
            self.stats_reader.close()
        if self.drain:
            timeout = timeouter.get_remaining_timeout()
            logger.info('Waiting for gun drain to finish for %f seconds', timeout)
            self.drain.join(timeout)
            if self.drain.is_alive():
                logger.warning('The gun drain didn\'t finish in time. Some data might be lost.')
            self.drain.close()

            timeout = timeouter.get_remaining_timeout()
            logger.info('Waiting for stats drain to finish for %f seconds', timeout)
            self.stats_drain.join(timeout)
            if self.drain.is_alive():
                logger.warning('The stats drain didn\'t finish in time. Some data might be lost.')
            self.stats_drain.close()
        logger.info('Collecting remaining data')
        self._collect_data(end=True)
        return retcode

    def add_result_listener(self, listener):
        self.listeners.append(listener)

    def __notify_listeners(self, data, stats):
        """ notify all listeners about aggregate data and stats """
        for listener in self.listeners:
            listener.on_aggregated_data(data, stats)


class _Timeouter:
    def __init__(self, total_timeout: float):
        self.total_timeout = total_timeout
        self.started = datetime.now()

    def get_remaining_timeout(self) -> float:
        return max(0.1, self.total_timeout - (datetime.now() - self.started).seconds)



from __future__ import print_function

import argparse

from datetime import datetime

import signal
from yandextank.contrib.netort.netort.data_manager import DataSession
import pandas as pd
import logging

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)


def string_to_df_microsec(data):
    # start_time = time.time()
    try:
        df = pd.read_csv(
            StringIO(data), sep='\t', names=phout_columns, na_values='', dtype=dtypes, quoting=QUOTE_NONE  # noqa
        )
    except CParserError as e:  # noqa
        logger.error(e.message)  # noqa
        logger.error('Incorrect phout data: %s', data)  # noqa
        return

    df['ts'] = (df['send_ts'] * 1e6 + df['interval_real']).astype(int)
    df['tag'] = df.tag.str.rsplit('#', 1, expand=True)[0]
    # logger.debug("Chunk decode time: %.2fms", (time.time() - start_time) * 1000)
    return df


def get_handler(data_session):
    def handler(signum, frame):
        data_session.interrupt()
    return handler


def get_uploader(data_session, column_mapping, overall_only=False):
    """
    :type data_session: DataSession
    """
    _router = {}
    _overall = {
        'interval_real': data_session.new_true_metric('interval_real overall', raw=False, aggregate=True, source='tank'),
        'connect_time': data_session.new_true_metric('connect_time overall', raw=False, aggregate=True, source='tank'),
        'send_time': data_session.new_true_metric('send_time overall', raw=False, aggregate=True, source='tank'),
        'latency': data_session.new_true_metric('latency overall', raw=False, aggregate=True, source='tank'),
        'receive_time': data_session.new_true_metric('receive_time overall', raw=False, aggregate=True, source='tank'),
        'interval_event': data_session.new_true_metric('interval_event overall', raw=False, aggregate=True, source='tank'),
        'net_code': data_session.new_event_metric('net_code overall', raw=False, aggregate=True, source='tank'),
        'proto_code': data_session.new_event_metric('proto_code overall', raw=False, aggregate=True, source='tank')}

    def get_router(tags):
        """
        :param tags:
        :return: {'%tag': {'%column_name': metric_object(name, group)}}
        """
        if set(tags) - set(_router.keys()):
            [_router.setdefault(tag,
                                {col_name: data_session.new_true_metric(name + '-' + tag,
                                                                        raw=False,
                                                                        aggregate=True)
                                 for col_name, name in column_mapping.items()} if not overall_only else {}
                                )
             for tag in tags]
        return _router

    def upload_overall(df):
        for col_name, metric in _overall.items():
            df['value'] = df[col_name]
            metric.put(df)

    def upload_df(df):
        router = get_router(df.tag.unique().tolist())
        if len(router) > 0:
            for tag, df_tagged in df.groupby('tag'):
                for col_name, metric in router[tag].items():
                    df_tagged['value'] = df_tagged[col_name]
                    metric.put(df_tagged)
        upload_overall(df)

    return upload_overall if overall_only else upload_df


def main():
    parser = argparse.ArgumentParser(description='Process phantom output.')
    parser.add_argument('phout', type=str, help='path to phantom output file')
    parser.add_argument('--url', type=str, default='https://test-back.luna.yandex-team.ru/')
    parser.add_argument('--name', type=str, help='test name', default=str(datetime.utcnow()))
    parser.add_argument('--db_name', type=str, help='ClickHouse database name', default='luna_test')
    args = parser.parse_args()

    clients = [
        {'type': 'luna', 'api_address': args.url, 'db_name': args.db_name},
        {'type': 'local_storage'}
    ]
    data_session = DataSession({'clients': clients})
    data_session.update_job({'name': args.name})
    print('Test name: %s' % args.name)

    col_map_aggr = {name: 'metric %s' % name for name in
                    ['interval_real', 'connect_time', 'send_time', 'latency',
                     'receive_time', 'interval_event']}
    uploader = get_uploader(data_session, col_map_aggr, True)

    signal.signal(signal.SIGINT, get_handler(data_session))

    with open(args.phout) as f:
        buffer = ''
        while True:
            parts = f.read(128*1024)
            try:
                chunk, new_buffer = parts.rsplit('\n', 1)
                chunk = buffer + chunk + '\n'
                buffer = new_buffer
            except ValueError:
                chunk = buffer + parts
                buffer = ''
            if len(chunk) > 0:
                df = string_to_df_microsec(chunk)
                uploader(df)
            else:
                break
    data_session.close()


if __name__ == '__main__':
    main()



"""
Monitoring stdout pipe reader. Read chunks from stdout and produce data frames
"""
import logging
import queue
import json

from ..Telegraf.decoder import decoder

logger = logging.getLogger(__name__)


class MonitoringReader(object):
    def __init__(self, source):
        self.buffer = []
        self.source = source
        self.finished = False
        self.prev_check = None

    def __iter__(self):
        while not self.finished:
            try:
                yield self._decode_agents_data(self.source.get_nowait())
            except queue.Empty:
                return

    def _decode_agents_data(self, block):
        """
        decode agents jsons, count diffs
        """
        collect = []
        if block:
            for chunk in block.split('\n'):
                try:
                    if chunk:
                        prepared_results = {}
                        jsn = json.loads(chunk)
                        for ts, values in jsn.items():
                            for key, value in values.items():
                                # key sample: diskio-sda1_io_time
                                # key_group sample: diskio
                                # key_name sample: io_time
                                try:
                                    key_group, key_name = key.split('_')[0].split('-')[0], '_'.join(key.split('_')[1:])
                                except:  # noqa: E722
                                    key_group, key_name = key.split('_')[0], '_'.join(key.split('_')[1:])
                                if key_group in decoder.diff_metrics:
                                    if key_name in decoder.diff_metrics[key_group]:
                                        decoded_key = decoder.find_common_names(
                                            key)
                                        if self.prev_check:
                                            try:
                                                value = jsn[ts][key] - \
                                                    self.prev_check[key]
                                            except KeyError:
                                                logger.debug(
                                                    'There is no diff value for metric %s.\n'
                                                    'Timestamp: %s. Is it initial data?', key, ts, exc_info=True)
                                                value = 0
                                            prepared_results[decoded_key] = value
                                    else:
                                        decoded_key = decoder.find_common_names(
                                            key)
                                        prepared_results[decoded_key] = value
                                else:
                                    decoded_key = decoder.find_common_names(
                                        key)
                                    prepared_results[decoded_key] = value
                            self.prev_check = jsn[ts]
                            collect.append((ts, prepared_results))
                except ValueError:
                    logger.exception('Telegraf agent send trash to output: %s', chunk)
                    return []
                except BaseException:
                    logger.exception('Exception trying to parse agent data: %s', chunk)
                    return []
            if collect:
                return collect



"""
Phantom phout format reader. Read chunks from phout and produce data frames
"""
from _csv import QUOTE_NONE

import pandas as pd
import numpy as np
import logging
import json
import time
import datetime
import itertools as itt

from pandas.errors import ParserError

from yandextank.common.interfaces import StatsReader

from io import StringIO

logger = logging.getLogger(__name__)

phout_columns = [
    'send_ts', 'tag', 'interval_real', 'connect_time', 'send_time', 'latency',
    'receive_time', 'interval_event', 'size_out', 'size_in', 'net_code',
    'proto_code'
]

dtypes = {
    'time': np.float64,
    'tag': np.str_,
    'interval_real': np.int64,
    'connect_time': np.int64,
    'send_time': np.int64,
    'latency': np.int64,
    'receive_time': np.int64,
    'interval_event': np.int64,
    'size_out': np.int64,
    'size_in': np.int64,
    'net_code': np.int64,
    'proto_code': np.int64,
}


def string_to_df(data):
    try:
        chunk = pd.read_csv(StringIO(data), sep='\t', names=phout_columns, dtype=dtypes, quoting=QUOTE_NONE)
    except ParserError as e:
        logger.error(str(e))
        logger.error('Incorrect phout data: {}'.format(data))
        return

    chunk['receive_ts'] = chunk.send_ts + chunk.interval_real / 1e6
    chunk['receive_sec'] = chunk.receive_ts.astype(np.int64)
    # TODO: consider configuration for the following:
    chunk['tag'] = chunk.tag.str.rsplit('#', n=1, expand=True)[0]
    chunk.set_index(['receive_sec'], inplace=True)
    return chunk


def string_to_df_microsec(data):
    # start_time = time.time()
    try:
        df = pd.read_csv(StringIO(data), sep='\t', names=phout_columns, na_values='', dtype=dtypes, quoting=QUOTE_NONE, float_precision="legacy")
    except ParserError as e:
        logger.error(str(e))
        logger.error('Incorrect phout data: {}'.format(data))
        return

    df['ts'] = (df['send_ts'] * 1e6 + df['interval_real']).astype(int)
    df['tag'] = df.tag.str.rsplit('#', n=1, expand=True)[0]
    # logger.debug("Chunk decode time: %.2fms", (time.time() - start_time) * 1000)
    return df


class PhantomReader(object):
    def __init__(self, fileobj, cache_size=1024 * 1024 * 50, parser=string_to_df):
        self.buffer = ""
        self.phout = fileobj
        self.cache_size = cache_size
        self.parser = parser

    def __iter__(self):
        return self

    def __next__(self):
        data = self.phout.read(self.cache_size)
        if data is None:
            raise StopIteration
        else:
            parts = data.rsplit('\n', 1)
            if len(parts) > 1:
                chunk = self.buffer + parts[0] + '\n'
                self.buffer = parts[1]
                return self.parser(chunk)
            else:
                self.buffer += parts[0]
                return None


class PhantomStatsReader(StatsReader):
    def __init__(self, filename, phantom_info, get_start_time=lambda: 0, cache_size=1024 * 1024 * 50):
        self.phantom_info = phantom_info
        self.stat_buffer = ""
        self.stat_filename = filename
        self.closed = False
        self.cache_size = cache_size
        self.get_start_time = get_start_time
        self._start_time = None

    def _decode_stat_data(self, chunk):
        """
        Return all items found in this chunk
        """
        for date_str, statistics in chunk.items():
            date_obj = datetime.datetime.strptime(
                date_str.split(".")[0], '%Y-%m-%d %H:%M:%S')
            chunk_date = int(time.mktime(date_obj.timetuple()))
            instances = 0
            for benchmark_name, benchmark in statistics.items():
                if not benchmark_name.startswith("benchmark_io"):
                    continue
                for method, meth_obj in benchmark.items():
                    if "mmtasks" in meth_obj:
                        instances += meth_obj["mmtasks"][2]

            offset = chunk_date - 1 - self.start_time
            reqps = 0
            if 0 <= offset < len(self.phantom_info.steps):
                reqps = self.phantom_info.steps[offset][0]
            yield self.stats_item(chunk_date - 1, instances, int(reqps))

    def _read_stat_data(self, stat_file):
        chunk = stat_file.read(self.cache_size)
        if chunk:
            self.stat_buffer += chunk
            parts = self.stat_buffer.rsplit('\n},', 1)
            if len(parts) > 1:
                ready_chunk = parts[0]
                self.stat_buffer = parts[1]
                return self._format_chunk(ready_chunk)
        else:
            self.stat_buffer += stat_file.readline()
        return None

    def _format_chunk(self, chunk):
        chunks = [json.loads('{%s}}' % s) for s in chunk.split('\n},')]
        return list(itt.chain(*(self._decode_stat_data(chunk) for chunk in chunks)))

    @property
    def start_time(self):
        if self._start_time is None:
            self._start_time = int(self.get_start_time())
        return 0 if self._start_time is None else self._start_time

    def __iter__(self):
        """
        Union buffer and chunk, split using '\n},',
        return splitted parts
        """
        with open(self.stat_filename, 'r') as stat_file:
            while not self.closed:
                yield self._read_stat_data(stat_file)
            # read end:
            data = self._read_stat_data(stat_file)
            while data:
                yield data
                data = self._read_stat_data(stat_file)
            # buffer is always included

    def close(self):
        self.closed = True



# -*- coding: UTF-8 -*-
import logging
from io import StringIO

import numpy as np
import pandas as pd
import queue as q

from yandextank.aggregator import TimeChopper
from yandextank.aggregator import aggregator as agg
from yandextank.aggregator.aggregator import DataPoller

logger = logging.getLogger(__name__)

KNOWN_EXC = {
    "java.net.NoRouteToHostException": 113,
    "java.net.ConnectException": 110,
    "java.net.BindException": 99,
    "java.net.PortUnreachableException": 101,
    "java.net.ProtocolException": 71,
    "java.net.SocketException": 32,
    "java.net.SocketTimeoutException": 110,
    "java.net.UnknownHostException": 14,
    "java.net.URISyntaxException": 22,
    "java.io.FileNotFoundException": 2,
    "java.io.IOException": 5,
    "java.io.EOFException": 104,
    "org.apache.http.conn.ConnectTimeoutException": 110,
    "org.apache.commons.net.MalformedServerReplyException": 71,
    "org.apache.http.NoHttpResponseException": 32,
    "java.io.InterruptedIOException": 32,
    "javax.net.ssl.SSLHandshakeException": 5,
}


def _exc_to_net(param1, success):
    """ translate http code to net code. if accertion failed, set net code to 314 """
    if len(param1) <= 3:
        # FIXME: we're unable to use better logic here, because we should support non-http codes
        # but, we should look for core.util.HTTP or some other common logic
        # here
        if success:
            return 0
        else:
            return 314

    exc = param1.split(' ')[-1]
    if exc in KNOWN_EXC:
        return KNOWN_EXC[exc]
    else:
        logger.warning("Unknown Java exception, consider adding it to dictionary: %s", param1)
        return 41


def _exc_to_http(param1):
    """ translate exception str to http code"""
    if len(param1) <= 3:
        try:
            int(param1)
        except BaseException:
            logger.error("JMeter wrote some strange data into codes column: %s", param1)
        else:
            return int(param1)

    exc = param1.split(' ')[-1]
    if exc in KNOWN_EXC:
        return 0
    else:
        logger.warning("Unknown Java exception. %s", param1)
        return 0


exc_to_net = np.vectorize(_exc_to_net)
exc_to_http = np.vectorize(_exc_to_http)

# phout_columns = [
#     'send_ts', 'tag', 'interval_real', 'connect_time', 'send_time', 'latency',
#     'receive_time', 'interval_event', 'size_out', 'size_in', 'net_code',
#     'proto_code'
# ]

jtl_columns = [
    'send_ts', 'interval_real', 'tag', 'retcode', 'success', 'size_in',
    'grpThreads', 'allThreads', 'latency', 'connect_time'
]
jtl_types = {
    'send_ts': np.int64,
    'interval_real': np.int64,
    'tag': np.str_,
    'retcode': np.str_,
    'success': np.bool_,
    'size_in': np.int64,
    'grpThreads': np.int64,
    'allThreads': np.int64,
    'latency': np.int64,
    'connect_time': np.float64,
}


def fix_latency(row):
    if row['latency'] < row['connect_time']:
        if row['interval_real'] < row['connect_time']:
            latency = 0
        else:
            latency = row['interval_real'] - row['connect_time']
    else:
        latency = row['latency'] - row['connect_time']
    return latency


# timeStamp,elapsed,label,responseCode,success,bytes,grpThreads,allThreads,Latency
def string_to_df(data):
    chunk = pd.read_csv(StringIO(data),
                        sep='\t',
                        names=jtl_columns, dtype=jtl_types,
                        keep_default_na=False,
                        on_bad_lines='warn')
    chunk["receive_ts"] = (chunk["send_ts"] + chunk['interval_real']) / 1000.0
    chunk['receive_sec'] = chunk["receive_ts"].astype(np.int64)
    chunk['interval_real'] = chunk["interval_real"] * 1000  # convert to µs
    chunk.set_index(['receive_sec'], inplace=True)
    chunk_length = len(chunk)
    chunk['connect_time'] = (chunk['connect_time'].fillna(0) * 1000).astype(np.int64)
    chunk['latency'] = chunk['latency'] * 1000
    chunk['latency'] = chunk.apply(fix_latency, axis=1)
    chunk['send_time'] = np.zeros(chunk_length)
    chunk['receive_time'] = chunk['interval_real'] - \
        chunk['latency'] - chunk['connect_time']
    chunk['interval_event'] = np.zeros(chunk_length)
    chunk['size_out'] = np.zeros(chunk_length).astype(int)
    chunk['net_code'] = exc_to_net(chunk['retcode'], chunk['success'])
    chunk['proto_code'] = exc_to_http(chunk['retcode'])
    return chunk


class JMeterStatAggregator(object):
    def __init__(self, source):
        self.worker = agg.Worker({"allThreads": ["max"]}, False)
        self.source = source

    def __iter__(self):
        for ts, chunk, rps in self.source:
            stats = self.worker.aggregate(chunk)
            yield [{
                'ts': ts,
                'metrics': {
                    'instances': stats['allThreads']['max'],
                    'reqps': rps
                }
            }]

    def close(self):
        pass


class JMeterReader(object):
    def __init__(self, filename, poller: DataPoller):
        self.buffer = ""
        self.stat_buffer = ""
        self.jtl_file = filename
        self.jmeter_finished = False
        self.agg_finished = False
        self.closed = False
        self.stat_queue = q.Queue()
        self.stats_reader = JMeterStatAggregator(
            TimeChopper([self._notify_poller_finished(poller.poll(self._read_stat_queue()))]))

    def _notify_poller_finished(self, poller):
        for item in poller:
            yield item
        self.agg_finished = True

    def _read_stat_queue(self):
        while not self.closed:
            # for _ in range(self.stat_queue.qsize()):
            try:
                yield self.stat_queue.get_nowait()
            except q.Empty:
                yield None

    def _read_jtl_chunk(self, jtl):
        data = jtl.read(1024 * 1024 * 10)
        if data:
            parts = data.rsplit('\n', 1)
            if len(parts) > 1:
                ready_chunk = self.buffer + parts[0] + '\n'
                self.buffer = parts[1]
                df = string_to_df(ready_chunk)
                self.stat_queue.put(df)
                return df
            else:
                self.buffer += parts[0]
        else:
            if self.jmeter_finished:
                self.agg_finished = True
            jtl.readline()
        return None

    def __iter__(self):
        with open(self.jtl_file, 'r') as jtl:
            while not self.closed:
                yield self._read_jtl_chunk(jtl)
            yield self._read_jtl_chunk(jtl)

    def close(self):
        self.closed = True



"""Manages your test data

* create the DataSession
* specify metrics you want to save
* specify the backends
* this module will collect your data and save them to specified backends
"""

# TODO: import only specific things that we really need to export
from .manager import *  # noqa
from .common.interfaces import MetricData   # noqa



import logging
import time
import threading as th
import multiprocessing as mp
from queue import Empty, Full

from ...stepper import StpdReader

logger = logging.getLogger(__name__)


class BFGBase(object):
    """
    A BFG load generator that manages multiple workers as processes and
    threads in each of them and feeds them with tasks
    """

    def __init__(self, gun, instances, stpd_filename, cached_stpd=False,
                 green_threads_per_instance=None):
        logger.info(
            """
BFG using stpd from {stpd_filename}
Instances: {instances}
Gun: {gun.__class__.__name__}
""".format(
                stpd_filename=stpd_filename,
                instances=instances,
                gun=gun, ))
        self.instances = int(instances)
        self.instance_counter = mp.Value('i')
        self.results = mp.Queue(16384)
        self.gun = gun
        self.gun.results = self.results
        self.quit = mp.Event()
        self.task_queue = mp.Queue(1024)
        self.cached_stpd = cached_stpd
        self.stpd_filename = stpd_filename
        self.pool = [
            mp.Process(target=self._worker) for _ in range(self.instances)
        ]
        self.feeder = th.Thread(target=self._feed, name="Feeder")
        self.feeder.daemon = True
        self.workers_finished = False
        self.start_time = None
        self.plan = None
        self.green_threads_per_instance = green_threads_per_instance

    def start(self):
        self.start_time = time.time()
        for process in self.pool:
            process.daemon = True
            process.start()
        self.feeder.start()

    def running(self):
        """
        True while there are alive workers out there. Tank
        will quit when this would become False
        """
        return not self.workers_finished

    def stop(self):
        """
        Say the workers to finish their jobs and quit.
        """
        self.quit.set()
        # yapf:disable
        while sorted([
                self.pool[i].is_alive()
                for i in range(len(self.pool))])[-1]:
            time.sleep(1)
        # yapf:enable
        try:
            while not self.task_queue.empty():
                self.task_queue.get(timeout=0.1)
            self.task_queue.close()
            self.feeder.join()
        except Exception as ex:
            logger.info(ex)

    def _feed(self):
        """
        A feeder that runs in distinct thread in main process.
        """
        self.plan = StpdReader(self.stpd_filename)
        if self.cached_stpd:
            self.plan = list(self.plan)
        for task in self.plan:
            if self.quit.is_set():
                logger.info("Stop feeding: gonna quit")
                return
            # try putting a task to a queue unless there is a quit flag
            # or all workers have exited
            while True:
                try:
                    self.task_queue.put(task, timeout=1)
                    break
                except Full:
                    if self.quit.is_set() or self.workers_finished:
                        return
                    else:
                        continue
        workers_count = self.instances
        logger.info("Feded all data. Publishing %d killer tasks" % workers_count)
        retry_delay = 1
        for _ in range(5):
            try:
                [
                    self.task_queue.put(None, timeout=1)
                    for _ in range(0, workers_count)
                ]
                break
            except Full:
                logger.debug(
                    "Couldn't post killer tasks"
                    " because queue is full. Retrying in %ss", retry_delay)
                time.sleep(retry_delay)
                retry_delay *= 2

        try:
            logger.info("Waiting for workers")
            for x in self.pool:
                x.join()
            logger.info("All workers exited.")
            self.workers_finished = True
        except (KeyboardInterrupt, SystemExit):
            self.task_queue.close()
            self.results.close()
            self.quit.set()
            logger.info("Going to quit. Waiting for workers")
            for x in self.pool:
                x.join()
            self.workers_finished = True


class BFGMultiprocessing(BFGBase):
    """
    Default worker type, creates process per worker,
    every process executes requests synchronously inside.
    """
    def _worker(self):
        """
        A worker that does actual jobs
        """
        logger.debug("Init shooter process")
        try:
            self.gun.setup()
        except Exception:
            logger.exception("Couldn't initialize gun. Exit shooter process")
            return
        while not self.quit.is_set():
            try:
                task = self.task_queue.get(timeout=1)
                if not task:
                    logger.debug("Got killer task.")
                    break
                timestamp, missile, marker = task
                planned_time = self.start_time + (timestamp / 1000.0)
                delay = planned_time - time.time()
                if delay > 0:
                    time.sleep(delay)

                try:
                    with self.instance_counter.get_lock():
                        self.instance_counter.value += 1
                    self.gun.shoot(missile.decode('utf8'), marker)
                finally:
                    with self.instance_counter.get_lock():
                        self.instance_counter.value -= 1

            except (KeyboardInterrupt, SystemExit):
                break
            except Empty:
                if self.quit.is_set():
                    logger.debug("Empty queue. Exiting process")
                    return
            except Full:
                logger.warning("Couldn't put to result queue because it's full")
            except Exception:
                logger.exception("Bfg shoot exception")

        try:
            self.gun.teardown()
        except Exception:
            logger.exception("Couldn't finalize gun. Exit shooter process")
            return
        logger.debug("Exit shooter process")


class BFGGreen(BFGBase):
    """
    Green version of the worker. Starts `self.instances` processes,
    each of process has a pool of `self.green_threads_per_instance` green threads.
    """
    def _worker(self):
        from gevent import monkey, spawn
        from gevent.queue import Queue as GreenQueue
        # NOTE: Patching everything will conflict with multiprocessing
        monkey.patch_all(thread=False, select=False)

        logger.debug("Init shooter process")
        try:
            self.gun.setup()
        except Exception:
            logger.exception("Couldn't initialize gun. Exit shooter process")
            return

        self.green_queue = GreenQueue(self.green_threads_per_instance)
        self.green_pool = [spawn(self._green_worker) for _ in range(0, self.green_threads_per_instance)]

        # Keep track of tasks sent to greenlets. If all greenlets are busy -
        # don't pull more tasks from the main queue, let other workers do that.
        self._free_threads_count = self.green_threads_per_instance

        while not self.quit.is_set():
            while not self.task_queue.empty() and self._free_threads_count:
                try:
                    task = self.task_queue.get_nowait()
                except Empty:
                    continue

                self._free_threads_count -= 1

                if not task:
                    logger.debug("Got killer task.")
                    self.quit.set()
                    break

                self.green_queue.put(task)

            time.sleep(0.1)

        for g in self.green_pool:
            g.join()

        try:
            self.gun.teardown()
        except Exception:
            logger.exception("Couldn't finalize gun. Exit shooter process")
            return
        logger.debug("Exit shooter process")

    def _green_worker(self):
        """
        A worker that does actual jobs
        """
        while not self.quit.is_set():
            try:
                task = self.green_queue.get(timeout=1)

                timestamp, missile, marker = task
                planned_time = self.start_time + (timestamp / 1000.0)
                delay = planned_time - time.time()

                if delay > 0:
                    time.sleep(delay)

                try:
                    with self.instance_counter.get_lock():
                        self.instance_counter.value += 1

                    self.gun.shoot(missile.decode('utf8'), marker)
                finally:
                    with self.instance_counter.get_lock():
                        self.instance_counter.value -= 1

                    self._free_threads_count += 1

            except (KeyboardInterrupt, SystemExit):
                break
            except Empty:
                continue
            except Full:
                logger.warning("Couldn't put to result queue because it's full")
            except Exception:
                logger.exception("Bfg shoot exception")
